/* C&C CA client-server server-only definitions */

#ifndef CASERVER_H
#define CASERVER_H

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include <string.h>
#include <openssl/bio.h>
#include <openssl/err.h>
#include <openssl/pem.h>
#include <openssl/rand.h>
#include <openssl/ssl.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <pthread.h>

#include "sslthread.h"

#include "ca.h"

#define CA_ROOT "/data/uwca/"

#define SV_CERT CA_ROOT "etc/sv_cert"

#define CRL_NAME  CA_ROOT "etc/crl.pem"
#define CA_CERT   CA_ROOT "etc/cacert.pem"
#define CA_KEY    CA_ROOT "etc/cacert.key"
#define CA_SERIAL CA_ROOT "db/serial"
#define CA_EXT_CS CA_ROOT "etc/config-cs"
#define CA_PID    CA_ROOT "db/caserver.pid"

#define CA_LOG    CA_ROOT "db/log"
#define CA_CERTDB CA_ROOT "db/certs"
#define CA_DEF_LIFETIME (60*60*24*365)
#define CA_MIN_LIFETIME (60*60*24*90)
#define CA_MAX_LIFETIME (60*60*24*366*5)


/* These are the db records.

   A zero casn value means a voided record. 
   An expire_date less that right now means an expired certificate.
   A non-zero revoke_date means a revoked certificate.

   The principal use of the db is to generate a CRL.

*/


typedef struct CertRec__ {
  long casn;            /* serial number */
  time_t issue_date;
  time_t expire_date;
  time_t revoke_date;
  char   dn[240];
} CertRec_, *CertRec;


#define THREAD_CC
#define THREAD_TYPE                    pthread_t
#define THREAD_CREATE(tid, entry, arg) pthread_create(&(tid), NULL, \
                                                      (entry), (arg))


#endif /* CASERVER_H */
