/* ca server command processing */

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>

#include "caserver.h"


/* read from a file into a string */

char *get_file(char *file, int *len)
{
   FILE *f = fopen(file,"r");
   char *str;
   long l;

   if (!f) return (NULL);
   fseek(f, 0, SEEK_END);
   l = ftell(f);
   rewind(f);

   str = (char*) malloc(l+1);
   fread(str, l, 1, f);
   str[l] = '\0';
   fclose(f);

   if (len) *len = l;
   return(str);
}



/* returns 0 on success */

ca_cmd(CmdHdr cmd, char *cmdtxt, RspHdr rsp, char **rsptxt)
{
  int ret = 0;
  char *resp = NULL;
  char *sout = NULL;
  char *pass = NULL;
   
  if (rsptxt) *rsptxt = NULL;
  if (*cmd->kpw) pass = cmd->kpw;

  switch (cmd->cmd) {

    case CA_SIGN:   /* sign a request */
       
       /* signer returns 0 on error */
       ret = ssl_sign(cmdtxt, cmd->ext, cmd->flags, pass, &resp);
       printf("Signed %d by %s\n", ret, cmd->id);
       break;
  
    case CA_REVOKE:   /* revoke a request  also does the gencrl */
       
       /* sn is in ext filed, dn is in cmdtxt */
       /* revoker returns 0 on error */
       ret = ssl_revoke(cmd->ext, cmdtxt);
       printf("Revoked %d by %s\n", ret, cmd->id);
       if (ret) ret = ssl_gencrl(pass, &resp);
       if (ret) resp = get_file(CRL_NAME, NULL);
       break;
  
    case CA_GENCRL:   /* gencrl - returns the new crl */
       
       ret = ssl_gencrl(pass, &resp);
       resp = get_file(CRL_NAME, NULL);
       if (resp) ret = 1;
       break;

    case CA_GETCRL:   /* getcrl */
       
       resp = get_file(CRL_NAME, NULL);
       if (resp) ret = 1;
       break;

    case CA_PKCS7:   /* sign a cert */
       
       /* signer returns 0 on error */
       ret = ssl_pkcs7(cmdtxt, cmd->flags, pass, &resp);
       printf("PKCS7 signed %d by %s\n", ret, cmd->id);
       break;
  
    default: ret = (0);

    }

    if (resp) {
       if (rsptxt) *rsptxt = resp;
       else free(resp);
    }

    if (ret) return (0);
    return (2);
  
}



