/* Sign a cert collection (PKCS7)  */

#include <sys/file.h>
#include <fcntl.h>

#include "caserver.h"

/* These are usually bad certs */
#define ERR(msg) {show_error(__FILE__, __LINE__, msg);pktxt=strdup(msg);goto err;}
/* These are internal and should never happen */
#define INTERR() ERR("INT");


/* Sign a cert (crttxt), returns 1, (pkcs7 in pktxt) */

int ssl_pkcs7(char *crttxt, int flags, char *keypass, char **pktxtp)
{
  char *pktxt = NULL;
  int i;
  EVP_PKEY *pkey, *CApkey;
  X509 *cert, *CAcert;
  X509_NAME *name;
  X509_NAME_ENTRY *nent;
  STACK_OF (X509) *certs;
  BIO *in, *out;
  FILE *fp;
  BUF_MEM *outbuf = NULL;
  int ret = 0;
  PKCS7 *pk;

  OpenSSL_add_all_algorithms ();
  ERR_load_crypto_strings ();
  seed_prng ();

  *pktxtp = NULL;
  
  /* ---  get the CA cert and key ------ */

  if (!(fp = fopen (CA_CERT, "r"))) INTERR ();
  if (!(CAcert = PEM_read_X509 (fp, NULL, NULL, NULL))) INTERR();
  fclose (fp);

  if (!(fp = fopen (CA_KEY, "r"))) INTERR();
  if (!(CApkey = PEM_read_PrivateKey (fp, NULL, NULL, keypass))) INTERR();
  fclose (fp);

  /* cert to sign is in crttxt */

  in = BIO_new_mem_buf(crttxt, strlen(crttxt));
  if (in == NULL) INTERR();

  BIO_set_close(in, BIO_NOCLOSE);

  cert = PEM_read_bio_X509(in,NULL,NULL,NULL);
  if (cert==NULL) {
     ERR("PKCS: bad cert");
     return (0);
  }
  // printf("crt dn = %s\n",ssl_get_dn(cert));

  certs = sk_X509_new_null();
  sk_X509_push(certs,cert);

  /* sign the certificate */
  pk = PKCS7_sign(CAcert, CApkey, certs, NULL, 0);
  if (pk==NULL) ERR("PKCS7: bad cert");

  /* return the signed pkcs7 */

  out = BIO_new(BIO_s_mem());
  if (out == NULL) INTERR();

  ret = BIO_get_mem_ptr(out, &outbuf);
  ret=PEM_write_bio_PKCS7(out, pk);

  pktxt = (char*) malloc(outbuf->length+1);
  memcpy(pktxt, outbuf->data, outbuf->length);
  pktxt[outbuf->length] = '\0';

  ret = 1;
 
/* Common bailout for errors */

err:

   /* Free all buffers and etc */

   *pktxtp = pktxt;

   return (ret);
}
