/* change the password on a key  */

#include <sys/file.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <curses.h>

#include <openssl/bio.h>
#include <openssl/err.h>
#include <openssl/pem.h>
#include <openssl/rand.h>
#include <openssl/ssl.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>


/* show the details of an error. These go to the local log. */
void show_error (const char *file, int lineno, const char *msg)
{
  fprintf (stderr, "** %s:%d %s\n", file, lineno, msg);
  ERR_print_errors_fp (stderr);
}

/* These are usually bad requests */
#define OERR(msg) {show_error(__FILE__, __LINE__, msg);exit(1);}



char *prog;
void usage()
{
  fprintf(stderr,"usage: %s key_file\n", prog);
  exit (1);
}



int main(int argc, char *argv[])
{
  int i;
  EVP_PKEY *pkey;
  const EVP_MD *digest;
  BIO *out;
  char *keyfile;
  char *outfile = NULL;
  int ret = 0;
  char opw[128];
  char npw1[128];
  char npw2[128];
  char *s;
  EVP_CIPHER *cipher=NULL;
  FILE *fp;

  prog = argv[0];

  OpenSSL_add_all_algorithms ();
  ERR_load_crypto_strings ();
  /* seed_prng (); */

  if (argc<2 || argc>3) usage();
  keyfile = argv[1];
  if (argc==3) outfile = argv[2];
  
  /* Read the key, using old password */

  fputs("Old password: ", stdout);
  noecho();
  fgets(opw,120,stdin);
  echo();
  if (s=strchr(opw,'\n')) *s = '\0';

  if (!(fp = fopen (keyfile, "r"))) OERR(keyfile);
  if (!(pkey = PEM_read_PrivateKey (fp, NULL, NULL, opw))) OERR("key");
  fclose (fp);

  /* Write the key, using new password */

  fputs("New password: ",stdout);
  noecho();
  fgets(npw1,120,stdin);
  echo();
  fputs("Again: ",stdout);
  noecho();
  fgets(npw2,120,stdin);
  echo();

  if (strcmp(npw1,npw2)) OERR("passwords dont match");
  if (s=strchr(npw1,'\n')) *s = '\0';
  if (npw1[0]) {
    cipher=EVP_des_ede3_cbc();
    if (strlen(npw1)<4) OERR("password needs at least 4 chars");
  } else {
    cipher=NULL;
  }
    
  out = BIO_new(BIO_s_file());
  if (out == NULL) OERR("bio");
  if (outfile) {
      if (BIO_write_filename(out,outfile)<=0) OERR(outfile);
  } else BIO_set_fp(out,stdout,BIO_NOCLOSE);

  if (!PEM_write_bio_PrivateKey(out,pkey,cipher,NULL,0,NULL,npw1)) OERR("out");

  exit (0);
}
