/* Revoke a request */

#include <sys/file.h>
#include <fcntl.h>
#include <errno.h>

#include "caserver.h"


/* log */

int write_log(int sn, char *msg)
{
  time_t now = time(NULL);
  struct tm *t = localtime(&now);
  int err = errno;
  FILE *log = fopen(CA_LOG,"a");

  if (log) {
     fprintf(log,"%4d/%02d/%02d %02d:%02d:%02d sn=%6d(%x): REV (%d) %s\n",
           t->tm_year+1900, t->tm_mon+1, t->tm_mday,
           t->tm_hour, t->tm_min, t->tm_sec,
           sn, sn, err, msg);
     fclose(log);
  }
}

/* Revoke a cert - just sets the revoke date in the DB
    returns sn if ok, 0 if error */

int ssl_revoke(int sn, char *dn)
{
  int crtdb = open(CA_CERTDB, O_CREAT|O_RDWR, 0660);
  CertRec C = (CertRec) malloc(sizeof(CertRec_));
  int nb;
  FILE *log;
  time_t now = time(NULL);

  if (!crtdb) return (0);

  if (lseek(crtdb, sn*sizeof(CertRec_), SEEK_SET)>0) {
     nb = read(crtdb, C, sizeof(CertRec_));
     if (nb!=sizeof(CertRec_)) {
        write_log(sn,"read");
        close(crtdb);
        return (0);
     }
     if (C->casn!=sn) {
        write_log(sn,"sn mismatch");
        close(crtdb);
        return (0);
     }
     if (!strstr(dn,C->dn)) {
        write_log(sn,"dn mismatch");
        fprintf(stderr, "revdn=%s, db=%s\n", dn, C->dn);
        close(crtdb);
        return (0);
     }
     C->revoke_date = now;
     if (lseek(crtdb, sn*sizeof(CertRec_), SEEK_SET)>0) {
        nb = write(crtdb, C, sizeof(CertRec_));
        if (nb!=sizeof(CertRec_)) {
           write_log(sn,"write");
        } else write_log(sn, "OK");
     } else write_log(sn, "wseek");
  } else write_log(sn, "rseek");
 
  close(crtdb);
  free (C);
  return (sn);
}


