/* CA utility provider: automatic expiration.

   Any certificates past expiration are set to EXPIRED.
   Owners are notified of soon-to-expire certificates.

   Requests whose certificates have expired more than 10 
   days ago are expired.
 
 */

#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <time.h>

#include "webtpl.h"

#include "db.h"
#include "util.h"
#include "cassl.h"
#include "ca.h"


extern int errno;

char *errstr = NULL;
static char *remote_user = "somebody";


main(int argc, char **argv)
{
  char *user;
  char *seqstr;
  int seq = 0;
  int s;
  char *p;
  char *msgstr = NULL;
  int reqno;

  SvrReq R = (SvrReq) malloc(sizeof(SvrReq_));
  SvrCert C = (SvrCert) malloc(sizeof(SvrCert_));

  X509 *xcrt;
  X509_REQ *xreq;
  int rc;
  char smsg[24];
  int okdn;
  time_t now = time(NULL);
  char tmsg[64];
  int maxdel = 20;     

  strcpy(tmsg, ctime(&now));
  if (p=strchr(tmsg,'\n')) *p = '\0';

  db_connect();

  /* Delete expired requests (and their certs and owners) */

  db_select_req(0, NULL, NULL, ST_EXPIRED, NULL);

  maxdel = 20;     
  while ((maxdel>0) && db_get_req(R)) {
     if (R->status!=ST_EXPIRED) continue;
     /* make sure the cert has expired */
     db_select_crt(R->crtno, R->reqno, 0);
     if (db_get_crt(C) && (C->status!=CRT_EXPIRED)) continue;
     db_free_crt();

     maxdel--;
     if (db_delete_req(R->reqno)) {
        printf("%s req %d deleted: user=%s, cn=%s\n", tmsg, R->reqno, R->reqid, R->dn);
        write_log(SERVICES_LOG, "somebody", "delreq", R->reqno, 0);
     } else {
        printf("%s req %d delete failed\n", tmsg, R->reqno);
     }
  }
  db_free_req();

  /* Delete orphaned certificates. */

  db_select_crt_orphans();

  maxdel = 20;
  while (((maxdel--)>0) && db_get_crt(C)>0) {
    if (db_delete_crt(C->crtno)) {
       printf("%s cert %d (sn=%d) deleted\n", tmsg, C->crtno, C->casn);
       write_log(SERVICES_LOG, "somebody", "delcrt", C->reqno, C->crtno);
    } else {
       printf("%s cert %d delete failed\n", tmsg, C->crtno);
    }
  }
  db_free_crt();

  /* Expire any certificates that have expired */

  db_select_crt_exp(0, 99);

  while (db_get_crt(C)>0) {
    if (db_set_crt_status(C->crtno, CRT_EXPIRED, C->reqno)) {
       printf("%s cert %d (sn=%d) expired\n", tmsg, C->crtno, C->casn);
       write_log(SERVICES_LOG, "somebody", "expcrt", C->reqno, C->crtno);
    } else {
       printf("%s cert %d expire failed\n", tmsg, C->crtno);
    }

  }
  db_free_crt();

  /* Notify users of soon to expire certs */

#ifdef DOINGEXPIRENOTICE
  db_select_crt_exp(MONTH, 0);

  while (db_get_crt(C)>0) {
    db_select_req(C->reqno, NULL, NULL, ST_NONE, NULL);
    if (db_get_req(R)) {
       char ssub[128];
       char *smsg = (char*)malloc(strlen(R->dn)+strlen(U_EXPIRE_MSG)+64);

       if (R->status != ST_ISSUE) continue;
       sprintf(ssub,"Certificate expiration warning: %s", R->dn);
       sprintf(smsg, U_EXPIRE_MSG, R->dn, C->casn, R->reqno);
       send_owner_mail(R, ssub, smsg);
       db_set_crt_expno(C->crtno, 1);
       printf("%s warn %s about %s\n", tmsg, R->reqid, R->dn);

       free(smsg);
    }
    db_free_req();
  }
  db_free_crt();
#endif

  /* Expire old requests (more than one month) */

  db_select_req_exp(-(MONTH));

  while (db_get_req(R)) {
     if (R->status==ST_RENEW) continue;  /* allow these to be signed */
     if (db_set_req_status(R->reqno, ST_EXPIRED, NULL)) {
        printf("%s req %d expired\n", tmsg, R->reqno);
        write_log(SERVICES_LOG, "somebody", "expreq", C->reqno, 0);
     } else {
        printf("%s req %d expire failed\n", tmsg, R->reqno);
     }
  }
  db_free_req();


}

