/* ca ssl tools */


#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#include <openssl/bio.h>
#include <openssl/err.h>
#include <openssl/ssl.h>
#include <openssl/buffer.h>
#include <openssl/objects.h>
#include <openssl/evp.h>
#include <openssl/rand.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/pem.h>
#include <openssl/pkcs12.h>
#include <openssl/des.h>


/* from x509 code */
/* Get a normal integer from an asn1 integer */

int i2i_ASN1_INTEGER(ASN1_INTEGER *a)
{
  int i;
  int h1, h2;
  int ret = 0;

  for (i=0;i<a->length;i++) {
     h1 = ((unsigned char)a->data[i]>>4)&0x0f;
     h2 = ((unsigned char)a->data[i]   )&0x0f;
     ret = ret*256 + h1*16 + h2;
  }
  return (ret);
}


/* ASN1 to time_t ( from coded-out part of a_utctm.c) */

#define g2(p) (((p)[0]-'0')*10+(p)[1]-'0')
time_t ASN1_UTCTIME_get(ASN1_UTCTIME *s)
{
   struct tm tm;
   int offset;

   memset(&tm,'\0',sizeof tm);

   tm.tm_year=g2(s->data);
   if(tm.tm_year < 50) tm.tm_year+=100;
   tm.tm_mon=g2(s->data+2)-1;
   tm.tm_mday=g2(s->data+4);
   tm.tm_hour=g2(s->data+6);
   tm.tm_min=g2(s->data+8);
   tm.tm_sec=g2(s->data+10);
   if(s->data[12] == 'Z') offset=0;
   else {
      offset=g2(s->data+13)*60+g2(s->data+15);
      if(s->data[12] == '-')
              offset= -offset;
   }
   return (mktime(&tm)-offset*60);
}
#undef g2

/* --------- end of openssl x509 extraction ----------- */

/* Get a request from a string */

X509_REQ *ssl_txt2req(char *reqstr)

{
  X509_REQ *req=NULL;
  BIO *rbio;
   
  rbio = BIO_new_mem_buf(reqstr, strlen(reqstr));
  if (rbio==NULL) {
    ERR_print_errors_fp(stderr);
    return (NULL);
  }

  if(!(req = PEM_read_bio_X509_REQ(rbio, NULL, NULL, NULL))) {
      /* fprintf(stderr, "Can't parse request %s\n", reqstr); */
      return (NULL);
  }

  /* close rbio */

  return (req);
}

#ifdef ALLOWBMPREWRITE

This is just some testing.

/* Rewrite a BMPSTRING subjectname to PRINTABLESTRING */

int ssl_rewrite_req_dn(X509_REQ *req) 
{
   X509_NAME *subj;
   X509_NAME_ENTRY *ne;
   int i,n,l,l1,j,l2;
   const char *s;
   int type, num;
   unsigned char *q;
   int gs_doit[4];
   char tmp_buf[80];
   static char hex[17]="0123456789ABCDEF";
   int need_rewrite = 0;
   int ret;
   char *p;
   char pbuf[16384];
   X509_NAME *nsubj;  
   int nid;

   subj = X509_REQ_get_subject_name(req);
   nsubj = X509_NAME_new();

   for (i=0; i<sk_X509_NAME_ENTRY_num(subj->entries); i++) {
      ne=sk_X509_NAME_ENTRY_value(subj->entries,i);
      nid=OBJ_obj2nid(ne->object);
      s=OBJ_nid2sn(nid);
      l1=strlen(s);
      type=ne->value->type;
      num=ne->value->length;
      q=ne->value->data;

      if ((type == V_ASN1_GENERALSTRING) && ((num%4) == 0)) {
               gs_doit[0]=gs_doit[1]=gs_doit[2]=gs_doit[3]=0;
               for (j=0; j<num; j++) if (q[j] != 0) gs_doit[j&3]=1;

               if (gs_doit[0]|gs_doit[1]|gs_doit[2]) gs_doit[0]=gs_doit[1]=gs_doit[2]=gs_doit[3]=1;
               else {
                       gs_doit[0]=gs_doit[1]=gs_doit[2]=0;
                       gs_doit[3]=1;
               }
      } else gs_doit[0]=gs_doit[1]=gs_doit[2]=gs_doit[3]=1;
      
       p = pbuf;
       for (j=0;j<num;j++) {
           n = q[j];
           if (n) *p++ = n;
           else need_rewrite = 1;
       }
       *p++ = 0;
       ret = X509_NAME_add_entry_by_NID(nsubj, nid, V_ASN1_PRINTABLESTRING, pbuf, strlen(pbuf), i, 0);
       printf(" add ret = %d\n");
  }

  ret = X509_REQ_set_subject_name(req, nsubj);
       printf(" set ret = %d\n");

   return (need_rewrite);
}
#endif /* allowrewrite */

/* Get the DN from a request */

char *ssl_get_dn_req(X509_REQ *req)

{
  X509_NAME *subj;
  char *dn;
  subj = X509_REQ_get_subject_name(req);
  dn =  X509_NAME_oneline(subj,NULL,0);
  return (dn);

}

/* Get the CN from a request */

char *ssl_get_cn_req(X509_REQ *req)

{
  X509_NAME *subj;
  int l;
  char *ret;

  subj = X509_REQ_get_subject_name(req);
  l = X509_NAME_get_text_by_NID(subj,NID_commonName,NULL,0);
  ret = (char*) malloc(l+8);
  X509_NAME_get_text_by_NID(subj,NID_commonName,ret,l+4);
  return (ret);
}


/* Get the alt names from a request */

#define MAX_ALT_NAMES 1000
char **ssl_get_altnames_req(X509_REQ *req)

{
  X509_NAME *subj;
  char **ret;
  char **rn;
  int nrn = 0;
  STACK_OF (X509_EXTENSION) *exts;
  int next;
  int i,l;

  exts = X509_REQ_get_extensions(req);
  if (!exts) return (NULL);
  next = sk_X509_EXTENSION_num(exts);

  ret = (char**) malloc((MAX_ALT_NAMES+1)*sizeof(char*));
  rn = ret;
  for (i=0; i<next; i++) {
     int nid;
     X509_EXTENSION *ext;
     ASN1_OBJECT *obj;

     ext = sk_X509_EXTENSION_value(exts, i);
     obj = X509_EXTENSION_get_object(ext);
     nid = OBJ_obj2nid(obj);
     if (nid == NID_subject_alt_name) {
        STACK_OF(GENERAL_NAME) *alt;
        GENERAL_NAME *an;

        alt = X509V3_EXT_d2i(ext);
        if (alt) {
           int j,anl;
           char *ant;
           for (j=0;j<sk_GENERAL_NAME_num(alt);j++) {
              an = sk_GENERAL_NAME_value(alt, j);
              if (an->type != GEN_DNS) continue;
              ant = ASN1_STRING_data(an->d.ia5);
              anl = ASN1_STRING_length(an->d.ia5);
              *rn++ = strdup(ant);
              if (nrn++==MAX_ALT_NAMES) break;
           }
        }
     }
  }
  *rn = NULL;
  return (ret);
}
 
/* Get the key usage flags from a request */

int ssl_get_keyusage_req(X509_REQ *req)

{
  STACK_OF (X509_EXTENSION) *exts;
  int next;
  int i,l;
  int ret = 0;

  exts = X509_REQ_get_extensions(req);
  if (!exts) return (0);
  next = sk_X509_EXTENSION_num(exts);

  for (i=0; i<next; i++) {
     int nid;
     X509_EXTENSION *ext;
     ASN1_OBJECT *obj;

     ext = sk_X509_EXTENSION_value(exts, i);
     obj = X509_EXTENSION_get_object(ext);
     nid = OBJ_obj2nid(obj);
     if (nid == NID_key_usage) {
        ASN1_BIT_STRING *usage;
        usage = X509V3_EXT_d2i(ext);
        // printf("key usage length = %d\n", usage->length);
        // printf("key usage 0 = %x\n", usage->data[0]);
        // printf("key usage 1 = %x\n", usage->data[1]);
        if (usage->length>0) ret = usage->data[0];
        if (usage->length>1) ret |= usage->data[1]<<8;
     }
     if (nid == NID_ext_key_usage) {
        ASN1_BIT_STRING *usage;
        usage = X509V3_EXT_d2i(ext);
        // printf("ext key usage length = %d\n", usage->length);
        // printf("ext key usage 0 = %x\n", usage->data[0]);
        // printf("ext key usage 1 = %x\n", usage->data[1]);
     }
  }
  return (ret);
}

/* See if this is a windows request */

#define WINDOWS_OID "1.3.6.1.4.1.311."
#define WINDOWS_OID_LEN 16

int ssl_is_windows_req(X509_REQ *req)
{
   STACK_OF(X509_ATTRIBUTE) *sk;
   X509_ATTRIBUTE *att;
   int i;
   char abuf[256];
   int l;

   sk = req->req_info->attributes;
   for (i=0;i<sk_X509_ATTRIBUTE_num(sk); i++) {
      att = sk_X509_ATTRIBUTE_value(sk,i);
      if (X509_REQ_extension_nid(OBJ_obj2nid(att->object))) continue;
      if ((l=i2t_ASN1_OBJECT(abuf, 256, att->object))>=WINDOWS_OID_LEN) {
         if (!strncmp(abuf, WINDOWS_OID, WINDOWS_OID_LEN)) return (1);
      }
   }
   return (0);
}   


/* Get a cert from a string */

X509 *ssl_txt2crt(char *crtstr)
{
  X509 *crt=NULL;
  BIO *cbio;

  cbio = BIO_new_mem_buf(crtstr, strlen(crtstr));
  if (cbio==NULL) {
    ERR_print_errors_fp(stderr);
    return (NULL);
  }

  if(!(crt = PEM_read_bio_X509(cbio, NULL, NULL, NULL))) {
      fprintf(stderr, "Can't parse cert %s\n", crtstr);
      return (NULL);
  }
  return (crt);
}


/* Get the DN from a cert */

char *ssl_get_dn(X509 *crt)
{
  X509_NAME *subj;
  subj = X509_get_subject_name(crt);
  return (X509_NAME_oneline(subj,NULL,0));
}


/* Get the CN from a cert */

char *ssl_get_cn(X509 *crt)
{
  X509_NAME *subj;
  subj = X509_get_subject_name(crt);
  int l = X509_NAME_get_text_by_NID(subj,NID_commonName,NULL,0);
  char *cn = (char*) malloc(l+8);
  *cn = '\0';
  X509_NAME_get_text_by_NID(subj,NID_commonName,cn,l+4);
  if (*cn) return (cn);
  free(cn);
  return (NULL);
}

/* Get the serial from a cert */

int ssl_get_sn(X509 *crt)
{
  X509_NAME *subj;
  int sn;
  sn = i2i_ASN1_INTEGER(crt->cert_info->serialNumber);
  return (sn);

}

/* Get the issue date from a cert */

time_t ssl_get_issue(X509 *crt)
{
  X509_NAME *subj;
  ASN1_UTCTIME *str;
  str = X509_get_notBefore(crt);
  return (ASN1_UTCTIME_get(str));
}


/* Get the expire date from a cert */

time_t ssl_get_expire(X509 *crt)
{
  X509_NAME *subj;
  ASN1_UTCTIME *expstr;
  expstr = X509_get_notAfter(crt);
  return (ASN1_UTCTIME_get(expstr));
}

 

/* Convert a cert to der encoding */

char *ssl_crttxt2der(char *crttxt, int *len)
{
  X509 *crt;
  BIO *dbio;
  int ret;
  BUF_MEM *derbuf;
  char *dertxt;

  crt = ssl_txt2crt(crttxt);
  if (!crt) return (NULL);
  dbio = BIO_new(BIO_s_mem());
  if (dbio==NULL) {
    ERR_print_errors_fp(stderr);
    return (NULL);
  }
  ret = BIO_get_mem_ptr(dbio, &derbuf);
  ret = i2d_X509_bio(dbio, crt);
  if (!ret) {
    ERR_print_errors_fp(stderr);
    return (NULL);
  }

  dertxt = (char*) malloc(derbuf->length);
  memcpy(dertxt, derbuf->data, derbuf->length);
  *len = derbuf->length;
  return (dertxt);
}


/* ------------------------------- */

