/* ca database routines (MySQL version) */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>

#include <mysql.h>

#include "webtpl.h"

#define UWCA_db_lib
#include "db.h"
#include "util.h"
#include "netact/netact.h"

extern char *remote_user;

static MYSQL mysql_db;
static MYSQL *mysql_certs = NULL;

static MYSQL_RES *req_res = NULL;
static MYSQL_RES *crt_res = NULL;
static MYSQL_RES *not_res = NULL;
static MYSQL_RES *usr_res = NULL;
static MYSQL_RES *own_res  = NULL;
static MYSQL_RES *xxx_res = NULL;

/* -------- DB open/close/error -------------------- */


/* Retrieve the last error from the DB */

char *db_last_error()
{
  MYSQL *m;
  m = mysql_certs;
  if (!m) m = &mysql_db;
  return ((char*)mysql_error(m));
}

void sql_err()
{
  char *e;
  MYSQL *m;
  if (1) return;   /* don't normally display errors */
  m = mysql_certs;
  if (!m) m = &mysql_db;
  e = (char*)mysql_error(m);
  if (!e) e = "-no msg-";
  printf("mysql: (%d) %s\n", mysql_errno(m), e);
}


/* Establish a db connection */

int db_connect() {
   FILE *fp;
   char *user = NULL;
   char *pass = NULL;
   char *host = NULL;
   char *db = "certs";
   char rec[128];
   char *p;

   if (mysql_certs) return(1);    /* already connected */

   if (!(fp=fopen(path_name(DB_CONFIG),"r"))) {
      printf("db config: %d", errno);
      return (0);
   }
   
   while (fgets(rec, 128, fp)) {
      if (p=strchr(rec,'\n')) *p = '\0';
      if (!strncmp(rec,"db=",3)) db = strdup(rec+3);
      if (!strncmp(rec,"user=",5)) user = strdup(rec+5);
      if (!strncmp(rec,"pass=",5)) pass = strdup(rec+5);
      if (!strncmp(rec,"host=",5)) host = strdup(rec+5);
   }
   fclose(fp);
      
   mysql_init(&mysql_db);
   mysql_options(&mysql_db, MYSQL_READ_DEFAULT_FILE, "/etc/my.cnf");
   
   mysql_certs = mysql_real_connect(&mysql_db, host, user, pass, db,
                                    0, NULL, 0);
   if (!mysql_certs) {
      sql_err();
      return (0);
   }
   return(1);
}

void db_disconnect()
{
   if (!mysql_certs) return;
   mysql_close(mysql_certs);
   mysql_certs = NULL;
}



/* ------------ DB data retrieval functions -------------- */


static time_t build_time(char *date)
{
    struct tm bldtm;
    int yr, mon, day;

    if (!date) return (0);

    memset (&bldtm,'\0', sizeof(bldtm));
    sscanf(date, "%d-%d-%d", &yr, &mon, &day);
    bldtm.tm_year = yr - 1900;
    bldtm.tm_mon = mon - 1;
    bldtm.tm_mday = day;
    bldtm.tm_isdst = (-1);
    return mktime(&bldtm);
}


/* make an sql style date */

static char *db_fix_date(time_t d)
{
   static char ret[12];
   struct tm *t;
   t = localtime(&d);
   sprintf(ret, "%04d-%02d-%02d", t->tm_year+1900, t->tm_mon+1, t->tm_mday);
   return(ret);
}



   
/* Retrieve next request record */

int db_get_req(SvrReq R)
{
   int rc;
   char *b;
   MYSQL_ROW row;

   if (!mysql_certs) return (0);

   if (!req_res) return (0);

   row = mysql_fetch_row(req_res);
   if (!row) {
      return (0); /* eof */
   }
 
   memset(R, '\0', sizeof(SvrReq_));
   R->reqno = atoi(row[0]);
   strcpy(R->reqid,row[1]);
   R->status = atoi(row[2]);
   R->request_date = build_time(row[3]);
   if (row[4]) R->crtno = atoi(row[4]);
   strcpy(R->request,row[5]);
   strcpy(R->dn,row[6]);
   if (row[7]) R->hash = atoi(row[7]);

   return (1);
}

/* Retrieve next certificate record */

int db_get_crt(SvrCert S)
{
   int rc;
   MYSQL_ROW row;

   if (!mysql_certs) return (0);

   if (!crt_res) return (0);

   row = mysql_fetch_row(crt_res);
   if (!row) {
      return (0); /* eof */
   }
 
   memset(S, '\0', sizeof(SvrCert_));
   S->crtno = atoi(row[0]);
   S->reqno = atoi(row[1]);
   S->casn = atoi(row[2]);
   S->status = atoi(row[3]);
   strcpy(S->cert,row[4]);
   S->issue_date = build_time(row[5]);
   if (row[6]) strcpy(S->issue_by,row[6]);
   S->revoke_date = build_time(row[7]);
   if (row[8]) strcpy(S->revoke_by,row[8]);
   S->expire_date = build_time(row[9]);
   if (row[10]) S->expno = atoi(row[10]);

   return (1);
}


/* Retrieve next notes record */

int db_get_notes(SvrNotes N)
{
   int rc;
   MYSQL_ROW row;

   if (!mysql_certs) return (0);

   if (!not_res) return (0);

   row = mysql_fetch_row(not_res);
   if (!row) {
      return (0); /* eof */
   }
 
   memset(N, '\0', sizeof(SvrNotes_));
   N->notno = atoi(row[0]);
   N->reqno = atoi(row[1]);
   strcpy(N->notes,row[2]);

   return (1);
}


/* Retrieve next user record */

int db_get_user(SvrUser U)
{
   int rc;
   MYSQL_ROW row;

   if (!mysql_certs) return (0);

   if (!usr_res) return (0);

   row = mysql_fetch_row(usr_res);
   if (!row) {
      return (0); /* eof */
   }
 
   memset(U, '\0', sizeof(SvrUser_));
   strcpy(U->id,row[0]);
   U->status = atoi(row[1]);

   return (1);
}

/* Retrieve next owner record */

int db_get_own(SvrOwn W)
{
   int rc;
   MYSQL_ROW row;

   if (!mysql_certs) return (0);

   if (!own_res) return (0);

   row = mysql_fetch_row(own_res);
   if (!row) {
      return (0); /* eof */
   }
 
   memset(W, '\0', sizeof(SvrOwn_));
   W->reqno = atoi(row[0]);
   strcpy(W->id,row[1]);

   return (1);
}



/* ------------ Retrieval selectors -------------- */


/* utility to assemble select's where clause */
static build_sel(char *sel, char **p)
{
   int f = 1;
   char *c = sel;

   while (*c) c++;
   while (*p) {
      if (f) strcat(strcpy(c, "where "), *p);
      else strcat(strcpy(c, " and "), *p);
      free (*p);
      while (*c) c++;
      p++;
      f = 0;
   }
   *c = '\0';
}


/* Select request records from the db.

   Can select by all, reqno, id, or status. */


int db_select_req(int reqno, char *dn, char *id, int status, char *srtkey)
{
   char cmd[1024];
   char sel[1024];
   char srt[1024];
   char *srtp;
   int rc;
   char *selp[5];
   int seli = 0;
   char *grp_join = "";;
   char *crt_join = "";;

   if (!mysql_certs) return (0);

   /* Build the select */

   if (reqno) {
      selp[seli] = (char*) malloc(48);
      sprintf(selp[seli], " svreq.reqno = %d ", reqno);
      seli++;
   }  
   if (dn && *dn) {
      selp[seli] = (char*) malloc(48);
      sprintf(selp[seli], " svreq.dn='%s' ", dn);
      seli++;
   }
   if (id && *id) {
      selp[seli] = (char*) malloc(48);
      sprintf(selp[seli], " svown.id='%s' ", id);
      seli++;
      grp_join = " left join svown on svreq.reqno=svown.reqno ";
   }
   if (status==RST_ACTIVE) {
      selp[seli] = (char*) malloc(96);
      sprintf(selp[seli],
            " (svreq.status=%d or svreq.status=%d or svreq.status=%d)",
            ST_REQ, ST_REVREQ, ST_RENEW);
      seli++;
   } else if (status==RST_ALIVE) {
      selp[seli] = (char*) malloc(96);
      sprintf(selp[seli], " (svreq.status=%d or svreq.status=%d)",
            ST_ISSUE, ST_RENEW);
      seli++;
   } else if (status==RST_CANBLE) {
      selp[seli] = (char*) malloc(96);
      sprintf(selp[seli], " (svreq.status=%d or svreq.status=%d)",
            ST_REQ, ST_RENEW);
      seli++;
   } else if (status==ST_NONE) {
      if (srtkey && srtkey[1]=='n') {
         selp[seli] = (char*) malloc(96);
         sprintf(selp[seli], " (svreq.status!=%d)", ST_REVOKED);
         seli++;
      }
   } else {
      selp[seli] = (char*) malloc(96);
      sprintf(selp[seli], " svreq.status = %d ", status);
      seli++;
   }
      
   selp[seli] = NULL;
   sel[0] = '\0';
   build_sel(sel, selp);

   /* Build sort */
  
   srtp = srt;
   while (srtkey && *srtkey) {
      int i = *srtkey++ - '0';
      if (!*srtkey) break;
      if (i<1||i>SR_NUM) {  /* not a sort key */
         srtkey++;
         continue;
      }
      if (srtp!=srt) *srtp++ = ',';
      strcpy(srtp, sort_vars[i]);
      if (*srtkey++=='d') strcat(srtp, " desc");
      while (*srtp) srtp++;
   }
   if (srtp==srt) strcpy(srt,"svreq.reqno");

   /* See if we need the svcrt join */

   if (strstr(srt, "svcrt.")) {
      crt_join = " left join svcrt on svreq.crtno=svcrt.crtno";
   }

   sprintf(cmd, "select * from svreq %s %s  %s order by %s",
              grp_join, crt_join, sel, srt);
   
   rc = mysql_query( mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }
   req_res = mysql_store_result(mysql_certs);
   if (!req_res) {
      sql_err();
      return (0);
   }
   
   return (1);
}


/* Select expired req records from the db. 
   
   Cert expired more than 'grace' seconds ago, or
   cert was revoked or revoke was requested. */
   

int db_select_req_exp(int grace)
{
   char cmd[1024];
   int rc;
   time_t t = time(NULL) + grace;

   if (!mysql_certs) return (0);

   sprintf(cmd,
   "select svreq.* from svreq,svcrt where svreq.crtno=svcrt.crtno and svreq.status!=%d and svcrt.status=%d and (expire_date<'%s' or svreq.status=%d or svreq.status=%d)",
             ST_EXPIRED, CRT_EXPIRED, db_fix_date(t), ST_REVOKED, ST_REVREQ);

   rc = mysql_query( mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }
   req_res = mysql_store_result(mysql_certs);
   if (!req_res) {
      sql_err();
      return (0);
   }
   return (1);
}



/* Select cert records from the db.
   
   Can select by crtno, reqno, and status. */


int db_select_crt(int crtno, int reqno, int status)
{
   char cmd[1024];
   char sel[256];
   int rc;
   char *selp[4];
   int seli = 0;

   if (!mysql_certs) return (0);

   /* Build the select */

   if (crtno) {
      selp[seli] = (char*) malloc(48);
      sprintf(selp[seli], " crtno = %d ", crtno);
      seli++;
   }
   if (reqno) {
      selp[seli] = (char*) malloc(48);
      sprintf(selp[seli], " reqno = %d ", reqno);
      seli++;
   }  
   if (status) {
      selp[seli] = (char*) malloc(48);
      sprintf(selp[seli], " status = %d ", status);
      seli++;
   }
   selp[seli] = NULL;
   sel[0] = '\0';
   build_sel(sel, selp);

   sprintf(cmd, "select * from svcrt %s", sel);
   
   rc = mysql_query( mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }
   crt_res = mysql_store_result(mysql_certs);
   if (!crt_res) {
      sql_err();
      return (0);
   }
   
   return (1);
}


/* Select expiring cert records from the db. */
   
int db_select_crt_exp(int grace, int lim)
{
   char cmd[1024];
   int rc;
   time_t t = time(NULL) + grace;

   if (!mysql_certs) return (0);

   sprintf(cmd,
   "select * from svcrt where status!=%d and expire_date<'%s' and expno<=%d",
             CRT_EXPIRED, db_fix_date(t), lim);

   rc = mysql_query( mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }
   crt_res = mysql_store_result(mysql_certs);
   if (!crt_res) {
      sql_err();
      return (0);
   }
   return (1);
}


/* Select orphaned cert records from the db. 
   Request points to renewed cert. */
   
int db_select_crt_orphans()
{
   char cmd[1024];
   int rc;

   if (!mysql_certs) return (0);

   sprintf(cmd,
      "select svcrt.* from svcrt,svreq where svcrt.status=%d and svcrt.reqno=svreq.reqno and svreq.crtno<>svcrt.crtno;", 
             CRT_EXPIRED);

   rc = mysql_query( mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }
   crt_res = mysql_store_result(mysql_certs);
   if (!crt_res) {
      sql_err();
      return (0);
   }
   return (1);
}



/* Select notes records from the db.
   
   Can select by reqno. */


int db_select_notes(int reqno)
{
   char cmd[1024];
   int rc;

   if (!mysql_certs) return (0);

   sprintf(cmd, "select * from svnot where reqno = %d order by notno", reqno);
   
   rc = mysql_query( mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }
   not_res = mysql_store_result(mysql_certs);
   if (!not_res) {
      sql_err();
      return (0);
   }
   return (1);
}


/* Select user records from the db. */
   

int db_select_user(char *id)
{
   char cmd[1024];
   int rc;

   if (!mysql_certs) return (0);

   if (id) sprintf(cmd, "select * from svusr where id = '%s'", id);
   else sprintf(cmd, "select * from svusr order by id");
   
   rc = mysql_query( mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }
   usr_res = mysql_store_result(mysql_certs);
   if (!usr_res) {
      sql_err();
      return (0);
   }
   return (1);
}



/* Select owner records from the db.
   
   Can select by userid, reqno. */


int db_select_own(int reqno, char *id)
{
   char cmd[1024];
   char sel[256];
   int rc;
   char *selp[4];
   int seli = 0;

   if (!mysql_certs) return (0);

   /* Build the select */

   if (id) {
      selp[seli] = (char*) malloc(48);
      sprintf(selp[seli], " id = '%s' ", id);
      seli++;
   }
   if (reqno) {
      selp[seli] = (char*) malloc(48);
      sprintf(selp[seli], " reqno = %d ", reqno);
      seli++;
   }  

   selp[seli] = NULL;
   sel[0] = '\0';
   build_sel(sel, selp);

   sprintf(cmd, "select * from svown %s", sel);
   
   rc = mysql_query( mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }
   own_res = mysql_store_result(mysql_certs);
   if (!own_res) {
      sql_err();
      return (0);
   }
   return (1);
}


/* ------------- Cursor maintenance ------------------ */


void db_free_req()
{
   if (req_res) mysql_free_result(req_res);
   req_res = NULL;
}

void db_free_crt()
{
   if (crt_res) mysql_free_result(crt_res);
   crt_res = NULL;
}

void db_free_notes()
{
   if (not_res) mysql_free_result(not_res);
   not_res = NULL;
}

void db_free_user()
{
   if (usr_res) mysql_free_result(usr_res);
   usr_res = NULL;
}

void db_free_own()
{
   if (own_res) mysql_free_result(own_res);
   own_res = NULL;
}

/* ------------ DB loader functions -------------- */


/* fix a string so it can be loaded into the db
   - escape quotes
    */

char *db_fix_string(char *cert)
{
  char *fcert;
  char *p;
  int nl = 2;
 
  /* malloc new string */
  for (fcert=cert;*fcert;fcert++,nl++) {
     if (*fcert=='\'') nl+=2;
  }
  fcert = (char*) malloc(nl);

  p = fcert;
  while (*cert) {
     if (*cert=='\'') *p++ = '\'';
     *p++ = *cert++;
   }
   *p++ = '\0';
   return (fcert);
}


/* Add a request record to the db.
   Returns the reqno of the new record or zero on error. */

int db_load_req(SvrReq R)
{
   char cmd[10000];
   char *fd, *fc, *fn;
   int rc, reqno=0;
   long hash;
   
   if (!mysql_certs) return (0);

   hash = fnv_hash_32(R->request);
   fd = db_fix_date(R->request_date);
   fc = R->request;
   fn = db_fix_string(R->dn);
   sprintf(cmd,
         "insert into svreq (reqid,status,request_date,request,dn,rhash) "
                " values ('%s',%d,'%s','%s','%s',%d)",
               R->reqid, R->status, 
               fd, fc, fn, hash);

   rc = mysql_query(mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }

   /* Get the request number */
   reqno = mysql_insert_id(mysql_certs);

   return (reqno);
}


/* Add a cert record to the db.
   Returns the crtno of the new record or zero on error. */

int db_load_crt(SvrCert S)
{
   char cmd[10000];
   char *fi, *fe, *fc;
   int rc, crtno=0;
   
   if (!mysql_certs) return (0);

   fi = strdup(db_fix_date(S->issue_date));
   fe = strdup(db_fix_date(S->expire_date));
   fc = S->cert;

   sprintf(cmd,
      "insert into svcrt " 
      " (reqno,status,casn,issue_date,issue_by,cert,expire_date,expno) "
                " values (%d,%d,%d,'%s','%s','%s','%s',0)",
               S->reqno, CRT_OK, S->casn, 
               fi, S->issue_by, fc, fe);

   rc = mysql_query(mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }

   /* Get the cert number */
   crtno = mysql_insert_id(mysql_certs);

   return (crtno);
}


/* Add a notes record to the db */

#define BOT2007 1167638400
int db_load_notes(int reqno, char *notes)
{
   char cmd[10000];
   char *fn;
   int rc;
   int nono = time(NULL) - BOT2007; /* just for sorting, really */
   
   if (!mysql_certs) return (0);

   fn = db_fix_string(notes);

   sprintf(cmd, "insert into svnot (notno,reqno,notes) values (%d,%d,'%s')",
               nono, reqno, fn); 

   rc = mysql_query(mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }

   return (nono);
}


/* Add a user record to the db */

int db_load_user(char *id, int status)
{
   char cmd[1024];
   int rc;
   
   if (!mysql_certs) return (0);

   sprintf(cmd, "insert into svusr (id,status) values ('%s',%d)",
               id, status); 

   rc = mysql_query(mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }

   return (1);
}


/* Add an owner record to the db */

int db_load_own(int reqno, char *id)
{
   char cmd[1024];
   int rc;
   
   if (!mysql_certs) return (0);

   sprintf(cmd, "insert into svown (reqno,id) values (%d,'%s')",
               reqno, id); 

   rc = mysql_query(mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }

   return (1);
}


/* ------------ Other DB data manipulation ------------------ */


/* update the status of a user record */

int db_set_user_status(char *id, int status)
{
   char cmd[1024];
   int rc;
   
   if (!mysql_certs) return (0);

   sprintf(cmd, "update svusr set status=%d where id = '%s'",
               status, id); 

   rc = mysql_query(mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }
   return (1);
}


/* update the crt number in a request record */

int db_set_req_crtno(int reqno, int crtno)
{
   char cmd[1000];
   int rc;
   
   if (!mysql_certs) return (0);

   sprintf(cmd, "update svreq set status=%d, crtno=%d where reqno = %d",
               ST_ISSUE, crtno, reqno); 

   rc = mysql_query(mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }
   return (1);
}



/* Set the status of a request record. Only some changes are allowed.
     
      REQ -> DENY
      REQ -> CANCEL
      DENY -> DELETED
      ISSUE -> RENEW
      ISSUE -> REVREQ
      REVREQ -> REVOKED
      any -> EXPIRED
   */

int db_set_req_status(int reqno, int status, char *id)
{
   char cmd[1024];
   char pssel[512];
   int ps = ST_NONE;
   int rc;
   
   if (!mysql_certs) return (0);

   if (status==ST_DENY) ps = ST_REQ;
   if (status==ST_DELETED) ps = ST_DENY;
   if (status==ST_RENEW) ps = ST_ISSUE;
   if (status==ST_CANCEL) ps = ST_REQ;
   /* if (status==ST_REVREQ) ps = ST_ISSUE; */
   if (status==ST_REVOKED) ps = ST_REVREQ;
   if (ps!=ST_NONE) sprintf(pssel,"and status=%d", ps);
   else pssel[0] = '\0';

   if (id) sprintf(cmd,
       "update svreq set status=%d,reqid='%s' where reqno=%d %s;",
               status, id, reqno, pssel); 
   else  sprintf(cmd,
       "update svreq set status=%d where reqno=%d %s;",
               status, reqno, pssel); 

   rc = mysql_query(mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }
   return (1);
}


/* Delete a request.  Must have current status REQ, DENY, or EXPIRED.
   Also deletes cert and owner records.  */
     
int db_delete_req(int reqno)
{
   char cmd[1024];
   int rc;
   
   if (!mysql_certs) return (0);

   /* delete any certificate records */
   sprintf(cmd, "delete from svcrt where reqno=%d and status=%d;", reqno, CRT_EXPIRED); 

   rc = mysql_query(mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }

   /* delete any ownership records */
   sprintf(cmd, "delete from svown where reqno=%d;", reqno); 

   rc = mysql_query(mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }

   /* delete the request */
   sprintf(cmd, "delete from svreq where reqno=%d "
      "and (status=%d or status=%d or status=%d);",
               reqno, ST_REQ, ST_DENY, ST_EXPIRED); 

   rc = mysql_query(mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }

   return (1);
}


/* Delete a certificate.  Must have current status EXPIRED. */
     
int db_delete_crt(int crtno)
{
   char cmd[1024];
   int rc;
   
   if (!mysql_certs) return (0);

   sprintf(cmd, "delete from svcrt where crtno=%d and status=%d;", crtno, CRT_EXPIRED); 

   rc = mysql_query(mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }

   return (1);
}


/* Set the status of a cert record. Only some changes are allowed.
   */

int db_set_crt_status(int crtno, int status, int reqno)
{
   char cmd[1024];
   int rc;
   
   if (!mysql_certs) return (0);

   sprintf(cmd,
    "update svcrt set status=%d where crtno=%d and reqno=%d;",
               status, crtno, reqno); 

   rc = mysql_query(mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }
   return (1);
}



/* Set the expno of a cert record. 
     
   */

int db_set_crt_expno(int crtno, int no)
{
   char cmd[1024];
   int rc;
   
   if (!mysql_certs) return (0);

   sprintf(cmd,
    "update svcrt set expno=%d where crtno=%d;",
               no, crtno); 

   rc = mysql_query(mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }
   return (1);
}


/* Set the revoke date of a cert record. 
   */

int db_set_crt_revoke_date(int crtno, time_t rdate)
{
   char cmd[1024];
   char *rd;
   int rc;
   
   if (!mysql_certs) return (0);

   rd = db_fix_date(rdate);
   sprintf(cmd,
    "update svcrt set revoke_date='%s' where crtno=%d and status=%d;",
               rd, crtno, CRT_REVOKED); 

   rc = mysql_query(mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }
   return (1);
}


     

/* Select request record by hash value.  Tests for duplicates.
   Returns 0 if request not found.
   */


int db_check_req_hash(long hash)
{
   char cmd[1024];
   int rc;
   MYSQL_ROW row;

   if (!mysql_certs) return (0);

   sprintf(cmd, "select reqno from svreq where rhash = %d and status != %d;", 
       hash, ST_CANCEL);
   
   rc = mysql_query(mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }
   
   xxx_res = mysql_store_result(mysql_certs);
   if (!xxx_res) {
      sql_err();
      return (0);
   }

   row = mysql_fetch_row(xxx_res);
   if (!row) {
      return (0); 
   }

   rc = atoi(row[0]);
   mysql_free_result(xxx_res);
   return (rc);
}


/* Get a user's status. 
   */

int db_get_user_status(char *id)
{
   SvrUser_ U;
   int s = US_UNKNOWN;

   if (!mysql_certs) db_connect();

   if (id && db_select_user(id)) {
      if (db_get_user(&U)) s = U.status;
      db_free_user();
   }
   return (s);
}



/* Delete a ownership record. */
     
int db_delete_own(int reqno, char *id)
{
   char cmd[1024];
   int rc;
   
   if (!mysql_certs) return (0);

   sprintf(cmd,
    "delete from svown where reqno=%d and id='%s'", reqno, id); 

   rc = mysql_query(mysql_certs, cmd);
   if (rc) {
      sql_err();
      return (0);
   }
   return (1);
}



/* ------------ Recon tools ----------------------- */
   
/* for debug output */
#define P if (0) printf

/* structs for owner recon */

typedef struct OwnerList__ {
   struct OwnerList__ *next;
   char *id;
   int status;
} OwnerList_, *OwnerList;

static OwnerList find_id_in_list(OwnerList L, char *id)
{
  OwnerList o;
  for (o=L; o; o=o->next) {
     if (!strcmp(o->id, id)) return (o);
  }
  return (NULL);
}

static OwnerList new_owner(OwnerList *root, char *id, int status, char *m)
{
   OwnerList o = (OwnerList) malloc(sizeof(OwnerList_));
   o->id = strdup(id);
   o->status = status;
   o->next = *root;
   *root = o;
  P(" add owner '%s' to '%s'\n", id, m);
   return (o);
}

static void free_owners(OwnerList O)
{
   OwnerList n;
   while (O) {
      n = O->next;
      if (O->id) free(O->id);
      free(O);
      O = n;
   }
}


/* Cache of domains */

typedef struct Domain__ {
   struct Domain__ *next;
   char *dn;
   OwnerList owners;
} Domain_, *Domain;

static Domain find_dn_in_list(Domain D, char *dn)
{
  Domain d;
  for (d=D; d; d=d->next) {
     if (!strcmp(d->dn, dn)) return (d);
  }
  return (NULL);
}

static Domain new_domain(Domain *root, char *dn)
{
   Domain d = (Domain) malloc(sizeof(Domain_));
   d->dn = strdup(dn);
   d->owners = NULL;
   d->next = *root;
   *root = d;
   P (" added domain '%s'\n", dn);
   return (d);
}

static void free_domains(Domain D)
{
   Domain n;
   while (D) {
      n = D->next;
      if (D->dn) free(D->dn);
      free(D);
      D = n;
   }
}


static void free_ids(char **ids)
{
   char **p;
   for (p=ids; p&&*p; p++) free(*p);
   free(ids);
}

/* Update the ownership records for a request */

static Domain dom_cache = NULL;

int db_update_owners(SvrReq R)
{
   SvrOwn_ W;
   char *dn = R->dn;
   OwnerList owners = NULL;
   OwnerList o, od;
   Domain d;
   char **ids, **idp;
   char *dp, *edp;
   int netact_ok = 1;
    
   if (!R->reqno) return (0);

   /* get existing owners from the DB */

   P ("dn=%s\n", dn);
   if (db_select_own(R->reqno, NULL)) {
      while (db_get_own(&W)) new_owner(&owners, W.id, 1, "R-db");
   }

   /* compare to netact host record */ 

   ids = netact_get_host_ids(dn);
   if (!ids) netact_ok = 0;     /* good connect always returns something */
   for (idp=ids; idp && *idp; idp++) {
      OwnerList o;
      if (o=find_id_in_list(owners, *idp)) {
         if (o->status==1) o->status = 2;
      } else new_owner(&owners, *idp, 3, "R-h");
   }
   free_ids(ids);

   /* check sub-domains */

   edp = strstr(dn, ".washington.edu");
   if (!edp) edp = strrchr(dn,'.');

   if (edp) {
      for (dp=strchr(dn,'.');dp&&dp<edp;dp=strchr(dp,'.')) {
          dp++;
          P ("  testing sub %s: ", dp);
          if (d=find_dn_in_list(dom_cache, dp)) {
            P ("  cached\n");
             for (od=d->owners;od;od=od->next) {
                P("   checking cached '%s'\n", od->id);
                if (o=find_id_in_list(owners, od->id)) {
                   if (o->status==1) o->status = 2;
                } else new_owner(&owners, od->id, 3, "R-sdc");
             }
          } else {
            P ("  new\n");
             d = new_domain(&dom_cache, dp);
             ids = netact_get_domain_ids(dp);
             if (!ids) netact_ok = 0;
             for (idp=ids; idp && *idp; idp++) {
                if (o=find_id_in_list(owners, *idp)) {
                    if (o->status==1) o->status = 2;
                } else new_owner(&owners, *idp, 3, "R-sdn");
                new_owner(&d->owners, *idp, 0, dn);
             }
             free_ids(ids);
          }
      }
   }

#ifdef KEEP_REQUESTER_AS_OWNER
   /* add the requester as owner by default */
   if (o=find_id_in_list(owners, R->reqid)) {
      if (o->status==1) o->status = 2;
   } else new_owner(&owners, R->reqid, 3, "R-req");
#endif
   
   /* Do necessary updates */

   if (netact_ok) {
      for (o=owners;o;o=o->next) {
         if (o->status==1) {
           db_delete_own(R->reqno, o->id);
           P (" recon delete '%s'\n", o->id);
         } else if (o->status==3) {
           db_load_own(R->reqno, o->id);
           P (" recon add '%s'\n", o->id);
         }
      }
   } else write_log(ERROR_LOG, remote_user, "netact", R->reqno, 0);

   free_owners(owners);

   return (1);

}
