/* ========================================================================
 * Copyright (c) 2004-2006 The University of Washington
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ========================================================================
 */


#include <stdio.h>
#include <math.h>

#include <ezp.h>

V x[360];
V ysin[360];
V ycos[360];
V ytan[360];
V ycot[360];
int nxy = 360;
int ntan = 85;

double twopi = 2.0 * 3.141592653589793;

void show_attrs(char *name, EZP_Graph_attrs ga)
{
  fprintf(stderr, "For %s:\n", name);
  fprintf(stderr, "    xp/yp = %d/%d\n", ga->xp, ga->yp);
  fprintf(stderr, "    xo/yo = %d/%g\n", ga->xo.i, ga->yo.f);
  fprintf(stderr, "    xs/ys = %g/%g\n", ga->xs, ga->ys);
}

main()
{
   EZP_Image ei;
   EZP_Graph g, gt, gc;
   EZP_Color bgc, fgc, sinc, cosc, tanc, cotc;
   EZP_Style s;
   int i;
   int xx = 0;
   EZP_Graph_attrs ga;

   ezp_verbose = 1;

   for (i=0;i<nxy;i++) {
      x[i].i = i;
      ysin[i].f = sin(i*twopi/360.0);
      ycos[i].f = cos(i*twopi/360.0);
   }
   for (i=0;i<ntan;i++) {
      ytan[i].f = tan(i*twopi/360.0);
   }
   for (i=5;i<ntan;i++) {
      ycot[i].f = 1.0/tan(i*twopi/360.0);
   }

   ei = ezp_create_image (900, 500);
   bgc = ezp_allocate_rgbcolor(ei, 255,255,255);
   fgc = ezp_allocate_rgbcolor(ei, 0,0,0);
   sinc = ezp_allocate_rgbcolor(ei, 255,0,0);
   cosc = ezp_allocate_rgbcolor(ei, 0,255,0);
   tanc = ezp_allocate_rgbcolor(ei, 100,100,0);
   cotc = ezp_allocate_rgbcolor(ei, 100, 0, 100);
   ezp_image_title(ei, "Sin, Cos, Tan and Cot", fgc);

   g = ezp_create_graph(ei, EZP_TYPE_INT, EZP_TYPE_FLOAT);
   ezp_graph_captions(g, "degree", "Sin & Cos", fgc);
   s = ezp_create_style(EZP_STYLE_LINE, 5, 5, sinc);
   ezp_plot_xy(g, x, ysin, nxy, s, "Sin");

   s = ezp_create_style(EZP_STYLE_LINE, 5, 5, cosc);
   ezp_plot_xy(g, x, ycos, nxy, s, "Cos");

   gt = ezp_create_graph(ei, EZP_TYPE_INT, EZP_TYPE_FLOAT);
   ezp_graph_captions(gt, "degree", "Tan", fgc);
   s = ezp_create_style(EZP_STYLE_FAT_LINE, 2, 4, tanc);
   ezp_plot_xy(gt, x, ytan, ntan, s, "Tan");

   gc = ezp_create_graph(ei, EZP_TYPE_INT, EZP_TYPE_FLOAT);
   ezp_graph_captions(gc, "degree", "Cot", fgc);
   s = ezp_create_style(EZP_STYLE_FAT_LINE, 2, 4, cotc);
   ezp_plot_xy(gc, x+5, ycot+5, ntan, s, "Cot");

   /* set size and position of the graphs */
   ezp_position_graph(g, 0.0, 0.0, 0.45, 1.0);
   ezp_position_graph(gt, 0.5, 0.0, 0.45, .5);
   ezp_position_graph(gc, 0.5, 0.4, 0.45, .5);

   ezp_save_image(ei,  EZP_IMAGE_JPG, "ezpdemo.jpg");

   ga = ezp_get_graph_attrs(g);
   show_attrs("sin/cos", ga);
   
   ga = ezp_get_graph_attrs(gt);
   show_attrs("tan", ga);

   ga = ezp_get_graph_attrs(gc);
   show_attrs("cot", ga);
   exit (0);
}
