/* ========================================================================
 * Copyright (c) 2004-2006 The University of Washington
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ========================================================================
 */

#include <stdio.h>
#include <math.h>
#include <string.h>

#include <ezp.h>

static char **xlabels;
static int nbar;
static int totx;
static V *zy, *zx;

typedef struct Hist__ {
   EZP_Graph g;
   int nbar;
   char **labels;
   V *x;
   V *y;
} Hist_, *Hist;
   

static char *hcb(V x, void *a)
{
   Hist H = (Hist) a;
   if (H->labels[x.i]) return (H->labels[x.i]);
   return ("");
}


Hist ezp_create_hist(EZP_Graph g, char **xl, int nx)
{
   Hist H = (Hist) malloc(sizeof(Hist_));
   int xp=0;
   int n,i;
   
   nbar = nx;
   totx = nx + 2;
   H->nbar = nbar;
   H->g = g;

   H->labels = (char**) malloc(sizeof(char*)*totx);
   H->x = (V*) malloc(sizeof(V)*totx);
   H->y = (V*) malloc(sizeof(V)*totx);
   
   H->labels[xp] = NULL;
   H->x[xp].i = xp++;

   for (n=0;n<nx;n++) {
      H->labels[xp] = strdup(xl[n]);
      H->x[xp].i = xp++;
   }

   H->labels[xp] = NULL;
   H->x[xp].i = xp++;

   ezp_set_graph_label_cb(g, hcb, NULL, H);

   return (H);
}

int ezp_plot_hist(Hist H, EZP_Style s, V *y, char *cap)
{
   int n;
   int i;
   int xp = 0;

   H->y[xp++].i = 0;

   for (n=0;n<H->nbar;n++) {
      H->y[xp++] = y[n];
   }

   H->y[xp++].i = 0;

   ezp_plot_xy(H->g, H->x, H->y, H->nbar+2, s, cap);

}


main()
{
   EZP_Image ei;
   EZP_Graph g, gt, gc;
   EZP_Color bgc, fgc, h1c, h2c, tanc;
   EZP_Style s;
   int i;
   int xx = 0;
   V h1[10];
   char *h1lbl[10];
   Hist h;


   ezp_verbose = 1;

   ei = ezp_create_image (900, 500);
   bgc = ezp_allocate_rgbcolor(ei, 255,255,255);
   fgc = ezp_allocate_rgbcolor(ei, 0,0,0);
   h1c = ezp_allocate_rgbcolor(ei, 255,0,0);
   h2c = ezp_allocate_rgbcolor(ei, 0,255,0);
   tanc = ezp_allocate_rgbcolor(ei, 100,100,0);
   ezp_image_title(ei, "Sin and Cos", fgc);

   h1lbl[0] = "h0";
   h1lbl[1] = "h1";
   h1lbl[2] = "h2";
   h1lbl[3] = "h3";
   h1lbl[4] = "h4";
   h1lbl[5] = "h5";


   g = ezp_create_graph(ei, EZP_TYPE_INT, EZP_TYPE_INT);
   ezp_graph_captions(g, "degree", " ", fgc);
   h = ezp_create_hist(g, h1lbl, 6);

   h1[0].i = 0;
   h1[1].i = 0;
   h1[2].i = 0;
   h1[3].i = 0;
   h1[4].i = 0;
   h1[5].i = 0;

   s = ezp_create_style(EZP_STYLE_BOX, 10, 10, bgc);
   ezp_plot_hist(h, s, h1, "");


   h1[0].i = 10;
   h1[1].i = 20;
   h1[2].i = 30;
   h1[3].i = 10;
   h1[4].i = 05;
   h1[5].i = 10;
  
   s = ezp_create_style(EZP_STYLE_BOX, 10, 10, h1c);
   ezp_plot_hist(h, s, h1, "h1s");

   h1[0].i = 12;
   h1[1].i = 25;
   h1[2].i = 35;
   h1[3].i = 12;
   h1[4].i = 00;
   h1[5].i = 19;

   s = ezp_create_style(EZP_STYLE_BOX, 10, 10, h2c);
   ezp_plot_hist(h, s, h1, "h2s");


   ezp_save_image(ei,  EZP_IMAGE_JPG, "hist.jpg");
   exit (0);
}
