/* ca tools */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <openssl/sha.h>

#include "webtpl.h"
#include "util.h"
#include "db.h"

extern char *remote_user; /* from user login */

/* browser and system type */


Browser get_browser()
{
   char *agent = getenv("HTTP_USER_AGENT");
   Browser B = (Browser) malloc(sizeof(Browser_));
   
   B->type = BR_XXX;
   B->version = 0;
   B->full_name = NULL;
   B->system = BS_XXX;

   if (agent) {
      B->full_name = strdup(agent);
      if (strstr(agent,"Opera")) B->type = BR_OPERA;
      else if (strstr(agent,"MSIE")) B->type = BR_IE;
      else if (strstr(agent,"Netscape")) B->type = BR_NS;
      else if (strstr(agent,"Chimera")) B->type = BR_CHIMERA;
      else if (strstr(agent,"OmniWeb")) B->type = BR_OMNIWEB;
      else if (strstr(agent,"Safari")) B->type = BR_SAFARI;
      else if (strstr(agent,"Gecko")) B->type = BR_GECKO;
   }

   if (agent) {
      if (strstr(agent,"Mac_Power")) B->system = BS_MAC;
      else if (strstr(agent,"Macintosh")) B->system = BS_MAC;
      else if (strstr(agent,"Windows")) B->system = BS_WIN;
      else if (strstr(agent,"Linux")) B->system = BS_LINUX;
   }

   return (B);
}


/* read from a file into a string */

char *get_file(char *file, int *len)
{
   FILE *f = fopen(file,"r");
   char *str;
   long l;

   if (!f) return (NULL);
   fseek(f, 0, SEEK_END);
   l = ftell(f);
   rewind(f);

   str = (char*) malloc(l+1);
   fread(str, l, 1, f);
   fclose(f);
   str[l] = '\0';

   if (len) *len = l;

   return(str);
}

/* write a string to a file */

int put_file(char *file, char *str)
{
   FILE *f = fopen(file, "w");
   fputs(str, f);
   return (1);
}

/* put a string to a temp file.  Returns the name
   of the file, or NULL if error. */

char *put_tmp_file(char *str, char *pfx)
{
   char *name = (char*) malloc(strlen(DIRROOT)+64);
   FILE *tf;
   sprintf(name,"%s/tmp/%s.%d",DIRROOT,pfx,getpid());
   unlink(name);
   tf = fopen(name, "w");
   if (tf) {
      fputs(str,tf);
      fclose(tf);
   }
   if (tf) return (name);
   else return (NULL);
}

/* fix a string for html display. newline -> br  */

char *fix_string_for_html(char *str)
{
  char *fstr;
  char *p;
  int nl = 5;

  if (!str) return ("");

  /* malloc new string */
  for (fstr=str;*fstr;fstr++,nl++) {
     if (*fstr=='\n') nl+=5;
     if (*fstr=='<') nl+=5;
     if (*fstr=='>') nl+=5;
  }
  fstr = (char*) malloc(nl);

  p = fstr;
  while (*str) {
     if (*str=='\n') {
        strcpy(p, "<br>\n");
        while (*p) p++;
     } else if (*str=='<') {
        strcpy(p, "&lt;");
        while (*p) p++;
     } else if (*str=='>') {
        strcpy(p, "&gt;");
        while (*p) p++;
     } else *p++ = *str;
     str++;
   }
   *p++ = '\0';
   return (fstr);
}

/* date strings */

char *date2mdy(time_t d)
{
   static char ret[12];
   struct tm *t;
   
   if (d) {
     t = localtime(&d);
     sprintf(ret, "%02d/%02d/%4d", t->tm_mon+1, t->tm_mday, t->tm_year+1900);
   } else strcpy(ret, "---");
   return(ret);
}

/* int strings */

char *int2dec(int i)
{
   static char ret[16];
   sprintf(ret, "%d", i);
   return(ret);
}

char *int2hex(int i)
{
   static char ret[16];
   sprintf(ret, "0x%x", i);
   return(ret);
}

/* cross site script provention */

/* This method generates a token unique to the input key.
   The token is added to forms and verified upon submission.
   We believe it is difficult to fake the token without
   knowing the encrypt seed (xsec_key). */

/* Get a token */

#ifdef OBSOLETE_XSEC
char *xsec_gen_token(char *key)
{
   int tk = time(NULL)/300;
   int i;
   static char tok[12];
   unsigned int v1=0;
   char *xsec_key;
   FILE *f;

   xsec_key = (char*) malloc(strlen(DIRROOT)+32);
   sprintf(xsec_key, "%s/etc/ca.key", DIRROOT);
   f = fopen(xsec_key, "r");
   if (!f) {
     printf("%s: vccould not open\n", xsec_key);
     return ("");
   }
   fscanf(f, "%6d", &v1);
   fclose(f);
   free(xsec_key);

   if (!key) key = getenv("REMOTE_USER");
   if (!key) key = getenv("REMOTE_ADDR");
   if (!key) key = "nokey";
   for (i=0;*key;key++,i++) v1+=(*key<<(i%24));

   srandom(v1);
   v1 = random() % 1000000000;
   sprintf(tok,"%0d", v1);
   return (tok);
}
   

/* Returns 0 if the token matches the key */

int xsec_chk_token(char *key, char *token)
{
   char *gk;
   if (!key) key = getenv("REMOTE_USER");
   if (!key) key = getenv("REMOTE_ADDR");
   if (!key) key = "nokey";
   gk = xsec_gen_token(key);
   return (strcmp(gk,token));
}

#endif

/* Generate a token - sha version 
   Token is sha1 hash of ( seed + userid(k1) + timestamp(k2) ) */

char *xsec_gen_token2(char *k2)
{
   unsigned char h[SHA_DIGEST_LENGTH];
   char *tok = (char*) malloc(SHA_DIGEST_LENGTH*2 + 1);
   unsigned char *k;
   char *k1 = remote_user;
   char *seed = get_file(XSEC_KEY_FILE, NULL);
   int kl;
   int i,j;

   if (!(seed&&k1&&k2)) return (NULL);  /* invalid use */

   k = (char*) malloc(strlen(seed)+strlen(k1)+strlen(k2)+8);
   strcat(strcat(strcpy(k,seed),k1),k2);
   kl = strlen(k);
   SHA1(k, kl, h);
   for (i=0,j=0;i<SHA_DIGEST_LENGTH;i++) {
      unsigned int x;
      tok[j++] = ((x=h[i]%16)<10)?x+'0':x-10+'a';
      tok[j++] = ((x=h[i]/16)<10)?x+'0':x-10+'a';
   }
   tok[j] = '\0';
   free (k);
   return (tok);
}

/* Returns 1 if the token matches the key */

int xsec_chk_token2(char *k2, char *token)
{
   char *gk;
   if (!(gk=xsec_gen_token2(k2))) return (0);
   return (strcmp(gk,token)==0);
}

/* Set xsec form fields */
void xsec_set_valid_form(WebTemplate W)
{
    time_t now = time(NULL);
    char nowtxt[24];
    sprintf(nowtxt,"%d",now);
    WebTemplate_assign(W, "XTIME", nowtxt);
    WebTemplate_assign(W, "XSITE", xsec_gen_token2(nowtxt));
}


/* See if the xsec tokens from the form are valid */

int xsec_chk_valid_form(WebTemplate W)
{
    char *xtime = WebTemplate_get_arg(W, "xtime");
    char *xsite = WebTemplate_get_arg(W, "xsite");
    if (!(xtime&&xsite&&*xtime&&*xsite)) {
       WebTemplate_assign(W, "EMSG", "Invalid form.");
       return (0);
    }
    time_t exp = atoi(xtime) + XSEC_FORM_TIMEOUT;
    time_t now = time(NULL);
    if (now>exp) {
       WebTemplate_assign(W, "EMSG", "Expired form.");
       return (0);
    }
    if (!xsec_chk_token2(xtime, xsite)) {
       WebTemplate_assign(W, "EMSG", "Xsite form.");
       return (0);
    }
    return (1);
}





/* Send mail */

static char *wday[] = {"Sun","Mon","Tue","Wed","Thu","Fri","Sat"};
static char *mon[] = {"Jan","Feb","Mar","Apr","May","Jun",
                      "Jul","Aug","Sep","Oct","Nov","Dec"};

int send_cc_mail(char *to, char *cc, char *sub, char *msg)
{
   FILE *sm = popen("/usr/sbin/sendmail -f " REPLY_TO " -t", "w");
   char dat[24];
   struct tm *tt;
   time_t t;
   int tz;

   t = time(NULL);
   tt = localtime(&t);
   tz = timezone/3600;
   if (tt->tm_isdst>0) tz -= 1;
   
   if (sm) {
      fprintf(sm,"Date: %s, %d %s %d %02d:%02d:%02d -%02d00\n",
                   wday[tt->tm_wday], tt->tm_mday, mon[tt->tm_mon],
                   tt->tm_year+1900, 
                   tt->tm_hour, tt->tm_min, tt->tm_sec, tz);
      fprintf(sm,"From: The CA <%s>\n", REPLY_TO);
      fprintf(sm,"To: %s\n", to);
      if (cc) fprintf(sm,"Cc: %s\n", cc);
      fprintf(sm,"Reply-To: %s\n", REPLY_TO); 
      fprintf(sm,"Errors-To: %s\n", REPLY_TO);
      fprintf(sm,"Subject: %s\n%s", sub, msg);
      fclose(sm);
      return (1);
   }
   return (0);
}

int send_mail(char *to, char *sub, char *msg)
{
   send_cc_mail(to, NULL, sub, msg);
}

/* u.washington.edu added to all ids */
int send_owner_mail(SvrReq R, char *sub, char *msg)
{
   SvrOwn_ O;
   char *cc = NULL;
   int cclen = 0;
   char own[48];
   
   /* make sure we have an up-to-date list */
   db_update_owners(R);

   db_select_own(R->reqno, NULL);
   while (db_get_own(&O)) {
      if (cc) {
         cclen += strlen(O.id)+24;
         cc = (char*) realloc(cc, cclen);
         strcat(cc,",");
         strcat(cc,O.id);
         strcat(cc,"@u.washington.edu");
      } else {
         cclen = 48;
         cc = (char*) malloc(48);
         strcpy(cc,O.id);
         strcat(cc,"@u.washington.edu");
      }
   }
   db_free_own();
   // Just send to the current owners
   // sprintf(own, "%s@u.washington.edu", R->reqid);
   send_cc_mail(cc, NULL, sub, msg);
   free(cc);
   return (1);
}

/* Make a full path from a relative one (static allocation) */

char *path_name(char *file)
{
   char *pb = (char*) malloc(strlen(DIRROOT)+strlen(file)+32);
   sprintf(pb, "%s/%s", DIRROOT, file);
   return (pb);
}

int write_log(char *name, char *user, char *action, int rno, int cno)
{
   FILE *logfile = fopen(path_name(name),"a");
   if (logfile) {
      time_t now = time(NULL);
      struct tm *t;
      t = localtime(&now);
      fprintf(logfile,
        "%04d/%02d/%02d %02d:%02d:%02d user=%s, action=%s, req=%d, crt=%d\n",
         t->tm_year+1900, t->tm_mon+1, t->tm_mday,
         t->tm_hour, t->tm_min, t->tm_sec,
          user, action, rno, cno);
      fclose(logfile);
      return (1);
    }
    return (0);
}

/* FNV hash for text of a request */

typedef unsigned long Fnv32_t;
typedef unsigned long long Fnv64_t;
#define FNV1_32_INIT ((Fnv32_t)0x811c9dc5)
#define FNV1_64_INIT ((Fnv64_t)0xcbf29ce484222325ULL)
#define FNV_32_PRIME ((Fnv32_t)0x01000193)
#define FNV_64_PRIME ((Fnv64_t)0x100000001b3ULL)


/* 32bit hash */
long fnv_hash_32(char *str)
{
    unsigned char *s = (unsigned char *)str;
    Fnv32_t hval = FNV1_32_INIT;

    while (*s) {
        if (isalnum(*s)) {
           hval *= FNV_32_PRIME;
           hval ^= (Fnv32_t)*s;
        }
        s++;
    }

    return (long)hval;
}

/* 16bit hash */
long long fnv_hash_64(char *str)
{
    unsigned char *s = (unsigned char *)str;
    Fnv64_t hval = FNV1_64_INIT;
    static char ret[17];

    while (*s) {
        if (isalnum(*s)) {
           hval *= FNV_64_PRIME;
           hval ^= (Fnv64_t)*s;
        }
        s++;
    }
    return (long long)hval;
}


/* convert a string to lower case */

char *lower_case(char *str)
{
  char *p;
  if (!str) return (NULL);
  for (p=str;*p;p++) if (isupper(*p)) *p = tolower(*p);
  return (str);
}


/* Add a crumb to a page */

void add_crumb(WebTemplate W, char *link, char *name)
{
  WebTemplate_assign(W, "CRUMBLINK", link);
  WebTemplate_assign(W, "CRUMBNAME", name);
  WebTemplate_parse_dynamic(W, "page.crumb.sep");
  WebTemplate_parse_dynamic(W, "page.crumb");
}

void crumb_init(WebTemplate W)
{
  WebTemplate_assign(W, "CRUMBLINK", "");
  WebTemplate_assign(W, "CRUMBNAME", "UW CA Home");
  WebTemplate_parse_dynamic(W, "page.crumb");
}

void crumb_help(WebTemplate W, char *link)
{
   WebTemplate_assign(W, "CRUMBLINK", link);
   WebTemplate_assign(W, "CRUMBNAME", "Help");
   WebTemplate_parse_dynamic(W, "page.crumbhelp");
}

void add_back_crumb(WebTemplate W, char *link, char *name)
{
  WebTemplate_assign(W, "CRUMBLINK", link);
  WebTemplate_assign(W, "CRUMBNAME", name);
  WebTemplate_parse_dynamic(W, "page.bcrumb.sep");
  WebTemplate_parse_dynamic(W, "page.bcrumb");
}


/* pack data - so it passes easily thru web pages */

static char hex[16] = "0123456789abcdef";
void util_pack(char **tgt, char *name, char *value)
{
   char *t = (char*) malloc(strlen(name) + strlen(value)*2 + 5);
   char *p = t;
   unsigned int v;
   unsigned int vh, vl;

   *p++ = 'z';
   strcpy(p, name);
   while (*p) p++;
   *p++ = 'x';
   while (v=*value++) {
      vh = v/16;
      vl = v - vh*16;
      *p++ = hex[vh];
      *p++ = hex[vl];
   }
   *p++ = 'z';
   *p = '\0';
   if (*tgt) {
      *tgt = (char*) realloc(*tgt, strlen(*tgt)+strlen(t)+5);
      strcat(*tgt, t);
   } else *tgt = t;
}

char *util_unpack(char *src, char *name)
{
   unsigned char *t;
   unsigned char *n;
   char *r;

   for (n=strstr(src, name); n; n=strstr(n, name)) {
      if ((*(n-1)=='z') && (*(t=(n+strlen(name)))=='x')) break;
   }
   if (!n) return (NULL);
   t++;
   if (!(n=strchr(t,'z'))) return (NULL);
   r = (char*) malloc(n-t);
   n = r; 
   while (*t && *t!='x') {
     unsigned int vh, vl;
     vh = *t - '0';
     if (vh<0||vh>9) vh = *t - 'a' + 10;
     if (!*++t || *t=='z') break;
     vl = *t - '0';
     if (vl<0||vl>9) vl = *t - 'a' + 10;
     *n++ = (char) vh*16+vl;
     t++;
   }
   *n = '\0';
   return (r);
}

/* maintain the sort key (cookie + form) 
   Sort key = <column_number><direction> */

char *set_sort(WebTemplate W, char *ckname)
{
   char *sk = (char*) malloc(SR_NUM*2+3);
   char *csk, *usk;
   time_t now = time(NULL);
   char *rsw;

   strcpy(sk, "ry2a3a");  /* default show and sort */

   /* Cookie holds old data */
   csk = WebTemplate_get_cookie(W, ckname);
   if (csk) {
      if (*csk=='r') strncpy(sk,csk,SR_NUM*2+2);
      else strncpy(sk+2,csk,SR_NUM*2);
   }
   sk[SR_NUM*2+3] = '\0';

   /* Show revoke arg is 'n' to hide. */

   rsw = WebTemplate_get_arg(W, "show_rev");
   if (rsw && *rsw) sk[1] = *rsw;

   /* Sort arg holds new sort column.  
      If new = old, then reverse order.
      Keep existing sort as secondard sort. */

   usk = WebTemplate_get_arg(W, "sort");
   if (usk) {
      char *nsk = (char*)malloc(SR_NUM*2+3);
      int i,j;
      /* save rsw part */
      nsk[0] = sk[0];
      nsk[1] = sk[1];
      /* see if this is swap */
      if (usk[0]==sk[2]) nsk[3] = (sk[3]=='a')?'d':'a';
      else nsk[3] = 'a';
      nsk[2] = usk[0];
      /* copy other keys */
      for (i=2,j=4;sk[i]&&i<SR_NUM*2+2;i+=2) {
        if (sk[i]!=usk[0]) nsk[j++] = sk[i], nsk[j++] = sk[i+1];
      }
      nsk[j] = '\0';
      free(sk);
      sk = nsk;
   }
   if (rsw||usk) {
      WebTemplate_set_cookie(W, ckname, sk,
         now+60*60*24*30*60, NULL,NULL,0);
   }
   if (!*sk) strcpy(sk,"ry2a3a");
   return (sk);
}


/* ------ interface to the auth system --- */

#ifdef DOING_LDAP_AUTH

#include "ldaplib.h"


#define CACRT "etc/ca_cert"
#define MYCRT "etc/uwa.ck"


int util_check_membership(char *grp, char *id, char **msg)
{
   LdapCfg cfg = new_ldap_cfg();
   void *L;
   int rc;
   char *err = NULL;

   cfg->host = "groups.u.washington.edu";
   cfg->port = 389;
   cfg->sslport = 389;

   cfg->certdb = path_name(CACRT);
   cfg->bindcrt = path_name(MYCRT);
   cfg->bindkey = path_name(MYCRT);

   L = uw_auth_initialize(cfg);

   if (!L) {
      if (err && msg) *msg = err;
      return (0);
   }

   rc = uw_auth_chk_in_group(L, cfg, id, grp);
   return (rc);
}

/* See if the user is 'special'.  Usually means can override limits */

#define SPECIAL_GROUP "u_cac_all"
int user_is_special(char *id)
{
   return (util_check_membership(SPECIAL_GROUP, id, NULL));
}

/* See if the user is an admin. */

/*
#define ADMIN_GROUP "Departments.C&C.Admin.CA admin"
int user_is_admin(char *id)
{
   return (util_check_membership(ADMIN_GROUP, id, NULL));
}
*/

#endif // DOING_LDAP_AUTH

#ifndef DOING_LDAP_AUTH
int user_is_special(char *id)
{
   if (strcmp(id,"uwit")) return (1);  // any local user
   return (0);
}
#endif
/* Possible lifetimes of a cert */

CertLife_ cert_lifetimes[] = {
  {1, "1 month"},
  {3, "3 months"},
  {6, "6 months"},
  {12, "1 year"},
  {24, "2 years"},
  {36, "3 years"},
  {0, NULL}
};

/* Compute lifetime in days of a new cert. 
   If this is a renewal count from old cert's expiration date.
   If no lifetime given use user-requested or default. */

int count_the_days(SvrReq R, int mon)
{
   SvrCert S = (SvrCert) malloc(sizeof(SvrCert_));
   SvrNotes N = (SvrNotes) malloc(sizeof(SvrNotes_));

   time_t now = time(NULL);
   time_t basetime = now;
   struct tm t;
   int yr = 0;
   int rmon = 0;

   if (R->crtno) {

      /* reference an existing good cert */
      db_select_crt(R->crtno, 0, CRT_OK);
      if (db_get_crt(S)) {
          basetime = S->expire_date;
          /* printf("using existing cert, base=%d\n", basetime); */
      } 
      db_free_crt(); 

   } 

      db_select_notes(R->reqno);
      while (db_get_notes(N)) {
         int rn, n;
         n = sscanf(N->notes, "Renewal of request %d by", &rn);
         if (n==1 && rn>1) {
            db_select_crt(0, rn, CRT_OK);
            if (db_get_crt(S)) {
               basetime = S->expire_date;
               /* printf("using reference from %d, base=%d\n", rn, basetime); */
            }
            db_free_crt();
         }
         n = sscanf(N->notes, "Requested lifetime: %d month", &rn);
         if (n==1 && rn>0) {
            /* printf("user requested %d months.\n", rn); */
            rmon = rn;
         }
      }
      db_free_notes(); 

   free(S);
   free(N);

   if (mon==0) mon = rmon;
   if (mon<=0) mon = DEF_CERT_LIFE;
   if (mon>MAX_CERT_LIFE) mon = MAX_CERT_LIFE;

   /* find day-offset from base time */

   /* printf("lifetime is %d months.\n", mon); */
   yr = mon/12;
   mon = mon - yr*12;
   localtime_r(&basetime, &t);
   t.tm_year += yr;
   t.tm_mon += mon;
   if (t.tm_mon>11) {
      t.tm_year++;
      t.tm_mon -= 12;
   }
   /* printf("expire is: %d/%d/%d %d:%d:%d \n", t.tm_mon+1, t.tm_mday, t.tm_year+1900, t.tm_hour, t.tm_min, t.tm_sec); */
   return (((mktime(&t)-now)/DAY)+1);
}

