
#ifndef _ca_util_h_
#define _ca_util_h_

#include <time.h>
#include "db.h"

/* browser and system type */

#define BR_XXX      0
#define BR_IE       1
#define BR_NS       2
#define BR_OPERA    3
#define BR_CHIMERA  4
#define BR_OMNIWEB  5
#define BR_SAFARI   6
#define BR_GECKO    7

#define BS_XXX      0
#define BS_LINUX    1
#define BS_MAC      2
#define BS_WIN      3

typedef struct Browser__ {
  int type;
  int version;
  char *full_name;
  int system;
} Browser_, *Browser;


/* Who gets mail notices */
#define GETS_MAIL "cert-dev@uw.edu"
#define REPLY_TO "iam-support@uw.edu"

#define SERVICES_LOG "log/services.log"
#define ERROR_LOG "log/error.log"
#define DB_CONFIG "etc/db.conf"

#define XSEC_FORM_TIMEOUT 3600
#define XSEC_KEY_FILE DIRROOT "/etc/xsec_key2"

/* Notice text */
#define U_NOTE_TXT \
 "\n\n** Tip ** : The UW Services CA manages certificate requests and sends email\n" \
 "notifications based on contact information in the Domain Name Service (DNS)\n" \
 "maintained by UW Information Technology. Requests for changes to the contact list\n" \
 "for your DNS name should be emailed to netops@uw.edu.\n"

/* Signed message */
#define U_SIGNED_MSG \
 "Your certificate request for %s (No. %d) has been signed. \n\n" \
 "Go to https://certs.cac.washington.edu/services/?req=full&rno=%d to retrieve it.\n\n" \
  U_NOTE_TXT

/* soon to expire message */
#define U_EXPIRE_MSG \
 "Your certificate for %s (sn=%d) will expire in one month.\n\n" \
 "If you intend to continue using this certificate\n" \
 "go to https://iam-tools.u.washington.edu/cs/ to renew it.\n\n" \
  U_NOTE_TXT

/* denied message */
#define U_DENIED_MSG \
 "Your certificate request for %s (No. %d) has been denied. \n\n" \
 "Go to https://certs.cac.washington.edu/services/?req=full&rno=%d to see why.\n\n" \
 U_NOTE_TXT

/* revoked message */
#define U_REVOKE_MSG \
 "Your request to revoke the certificate \n" \
 "for %s (No. %d) has been completed.\n\n" \
 U_NOTE_TXT


/* Cert lifetime selections */

#define DAY (24*60*60)
#define MONTH (30*24*60*60)
#define MAX_CERT_LIFE 36
#define DEF_CERT_LIFE 36

typedef struct CertLife__ {
  int mon;
  char *txt;
} CertLife_, *CertLife;


/* prototypes */
Browser get_browser();
char *get_file(char *file, int *len);
int put_file(char *file, char *str);
char *put_tmp_file(char *str, char *pfx);
char *date2mdy(time_t date);
char *int2dec(int i);
char *int2hex(int i);

char *xsec_gen_token(char *key);
int xsec_chk_token(char *key, char *token);
char *fix_string_for_html(char *str);

int send_cc_mail(char *to, char *cc, char *sub, char *msg);
int send_mail(char *to, char *sub, char *msg);
int send_owner_mail(SvrReq R, char *sub, char *msg);
int write_log(char *name, char *user, char *action, int rno, int cno);

long fnv_hash_32(char *str);
long long fnv_hash_64(char *str);

char *lower_case(char *str);
char *path_name(char *file);

void util_pack(char **tgt, char *name, char *value);
char *util_unpack(char *src, char *name);

char *set_sort(WebTemplate W, char *ckname);

int util_check_membership(char *grp, char *id, char **msg);
int user_is_special(char *id);

extern CertLife_ cert_lifetimes[];
int count_the_days(SvrReq R, int mon);



#endif
