/* CA utility provider: service cert page and request form  */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>

#include "webtpl.h"
#include "db.h"
#include "util.h"
#include "cassl.h"

extern char *remote_user;
extern int errno;
char *errstr = NULL;

#define CHECK "checked"
#define SELECT "selected"

#define REQ(s) (req && !strcmp(req,s))
#define RREQ(s) (req && !strcmp(req,s) && (reqno>0))
#define CREQ(s) (req && !strcmp(req,s) && (crtno>0))

#define ADMIN_URL \
    " Go to https://certs.cac.washington.edu/services/admin/ to handle it."
#define BADDN_FILE "baddn.txt"

/* handle client disconnect */

void catch_pipe(int a)
{
   signal(SIGPIPE, catch_pipe);
}

/* from request parameters */

int reqno;
int crtno;


/* xsfd check and exit */

void chk_xsec(WebTemplate W) {
   if (xsec_chk_valid_form(W)) return;
   WebTemplate_write(W, "PAGE");
   exit(1);
}

/* look for some text */

int has_some_text(char *str)
{
   for (;str&&*str;str++) if (isalnum(*str)) return (1);
   return (0);
}


/* Report a database error */

static void show_db_error(WebTemplate W, char *op)
{
   char *e = (char*) db_last_error();
   char *m = (char*)malloc(strlen(e)+strlen(op)+36);
   sprintf(m, "Internal DB error (%s): %s\n", op, e);
   WebTemplate_assign(W, "EMSG", m);
   send_mail("fox@washington.edu", "CA DB error", m);
   free(m);
}
   

/* See if a request is for a windows cert */

int is_windows_req(SvrReq R)
{
   X509_REQ *xreq = ssl_txt2req(R->request);
   return (ssl_is_windows_req(xreq));
}

/* add a crumb for a particular request */
void add_req_crumb(WebTemplate W, int reqno)
{
   char pt[64];
   char pl[64];
   sprintf(pl, "services/?req=full&rno=%d", reqno);
   sprintf(pt, "Request %d", reqno);
   add_crumb(W, pl, pt);
}

/* show the status of a request action */
void show_action_status(WebTemplate W, int reqno, char *status_key)
{
   char dyn[256];
   add_crumb(W, "services/?req=list", "Manage certificates");
   if (reqno) add_req_crumb(W, reqno);
   if (strlen(status_key)<250) {
      strcpy(dyn, "page.");
      strcat(dyn, status_key);
      WebTemplate_parse_dynamic(W, dyn);
   }
}

/* Fix a pem cert for inclusion in a vbs script */

char *fix_cert_for_vbs(char *crt)
{
   char *fix = (char*)malloc(strlen(crt)+1024);
   char *p = fix;
   char *cp;
   for (cp=crt;cp;) {
     *p++ = '"';
     cp = strchr(crt,'\n');
     if (cp) *cp++ = '\0';
     while (*crt) *p++ = *crt++;
     strcpy(p, "\" & _\n");
     p += 6;
     *p = '\0';
     crt = cp;
     if (!*cp) break;
   }
   p -= 5;
   *p = '\0';
   return (fix);
}

/* Check a request for valid dns names. 
   Returns NULL on valid, error message otherwise.  */

char *req_chk_dns(X509_REQ *xreq)
{
   char *err;
   char **alt;
   char *em;
   if (err=ssl_chk_dns(ssl_get_cn_req(xreq),0)) {
      em = (char*) malloc(strlen(err)+128);
      sprintf(em, "We cannot process the request.<br>%s", err);
      return (em);
   }
   for (alt=ssl_get_altnames_req(xreq);alt&&*alt;alt++) {
      if (err=ssl_chk_dns(*alt, 0)) {
         em = (char*) malloc(strlen(err)+128);
         sprintf(em, "We cannot process the request.<br>%s (subjectAltName)",
                            err);
         return (em);
      }
   }
   return (NULL);
}

/* Check a request for policy compliance.
   Returns NULL on valid, error message otherwise.  */

char *req_chk_policy(X509_REQ *xreq, char *id)
{
   char *err;
   char **alt;
   if (err=ssl_chk_req_dn(xreq, id)) return (err);
   if (err = ssl_chk_req_altnames(xreq, id)) {
      strncpy(err, "AN", 2);
      return (err);
   }
   return (NULL);
}

int show_req_detail(WebTemplate W, int reqno, int show_chks)
{
   SvrReq R = (SvrReq) malloc(sizeof(SvrReq_));
   SvrCert S = (SvrCert) malloc(sizeof(SvrCert_));
   SvrNotes N = (SvrNotes) malloc(sizeof(SvrNotes_));
   char *p, *dn, *cn;
   X509_REQ *xreq;
   char *ckemsg = NULL;
   int okdn = 0;
   int showrev = 0;
   int showren = 0;
   time_t now = time(NULL);
   char **alt;
   char *errdn = NULL;
   int ret;
   int life = 0;
   CertLife cl;

   if (!db_connect()) show_db_error(W, "connect");
    
   /* Make a quick ownership update */
   db_select_req(reqno, NULL, NULL, ST_NONE, NULL);
   if (db_get_req(R)>0) {
      db_update_owners(R);
      errdn = strdup(R->dn);
   }

   /* Ok get for this user */

   db_select_req(reqno, NULL, NULL, ST_NONE, NULL);
   if (db_get_req(R)>0) {

      xreq = ssl_txt2req(R->request);

      dn = ssl_get_dn_req(xreq);
      if (*dn=='/') dn++;
      // if (p=strchr(dn,'/')) *p++ = '\0';
      WebTemplate_assign(W, "DN1", dn);
/**
      while (p && *p) {
         if (dn=strchr(p,'/')) *dn++ = '\0';
         WebTemplate_assign(W, "DNX", p);
         WebTemplate_parse_dynamic(W, "reqfull.dn");
         p = dn;
      }
 **/
      for (alt=ssl_get_altnames_req(xreq);alt&&*alt;alt++) {
         WebTemplate_assign(W, "ALTNAME", *alt);
         WebTemplate_parse_dynamic(W, "reqfull.altname");
      }
      WebTemplate_assign(W, "DREQNO", int2dec(R->reqno));
      WebTemplate_assign(W, "DRDATE", date2mdy(R->request_date));
      WebTemplate_assign(W, "DUSER", R->reqid);
      WebTemplate_assign(W, "DSTATUS", req_status_text[R->status]);
        
      db_select_crt(R->crtno, R->reqno, 0);
      while (db_get_crt(S)) {
         WebTemplate_assign(W, "CASN", int2dec(S->casn));
         WebTemplate_assign(W, "CASNX", int2hex(S->casn));
         WebTemplate_assign(W, "SDATE", date2mdy(S->issue_date));
         WebTemplate_assign(W, "SIGNER", S->issue_by);
         WebTemplate_assign(W, "EDATE", date2mdy(S->expire_date));

         if ((R->status!=ST_REVREQ) && 
             (S->status==CRT_OK) && !show_chks) {
            WebTemplate_assign(W, "CRT", int2dec(S->crtno));
            if (is_windows_req(R)) {
               Browser B = get_browser();
               if (B->type==BR_IE) WebTemplate_parse_dynamic(W, "reqfull.cert.get.getms");
               else WebTemplate_parse_dynamic(W, "reqfull.cert.get.getmsnot");
            }
            WebTemplate_parse_dynamic(W, "reqfull.cert.get");
            showrev = 1;
         }
         if (R->status==ST_ISSUE && S->status!=CRT_REVOKED &&
             S->expire_date<now+MONTH*2) showren = 1;

         WebTemplate_parse_dynamic(W, "reqfull.cert");
      }
      db_free_crt();

      db_select_notes(R->reqno);
      while (db_get_notes(N)) {
         int n, rn;
         WebTemplate_assign(W, "DNOTE", fix_string_for_html(N->notes));
         WebTemplate_parse_dynamic(W, "reqfull.notes");
         n = sscanf(N->notes, "Requested lifetime: %d month", &rn);
         if (n==1 && rn>0) {
            life = rn;
         }
      }
      db_free_notes();
      if (life==0) life = DEF_CERT_LIFE;

      /* Check for policy compliance */
      ckemsg = req_chk_policy(xreq, R->reqid);
      okdn = ckemsg==NULL;
      if (!okdn) {
         WebTemplate_assign(W, "ERRMSG", ckemsg+4);
         WebTemplate_assign(W,"CN",ssl_get_cn_req(xreq));
         WebTemplate_parse_dynamic(W, "reqfull.baddn");
      }


      if (showren) {
         Browser B = get_browser();
         WebTemplate_assign(W,"CNX",R->dn);
         if (is_windows_req(R) && B->type!=BR_IE) WebTemplate_parse_dynamic(W, "reqfull.cant_renew");
         else {
            if (is_windows_req(R)) {
               /* add capicom test to the page */
               WebTemplate_get_by_name(W, "iebody", "ms/iebody.tpl");
               WebTemplate_get_by_name(W, "captest", "ms/iecaptest.tpl");
               WebTemplate_parse(W, "BODYTXT", "iebody");
               WebTemplate_parse(W, "IECAPTEST", "captest");
            }
            for (cl=cert_lifetimes; cl->mon; cl++) {
               WebTemplate_assign_int(W, "LMON", cl->mon);
               WebTemplate_assign(W, "LTXT", cl->txt);
               WebTemplate_assign(W, "LSEL", (life==cl->mon)?SELECT:"");
               WebTemplate_parse_dynamic(W, "reqfull.renew.lifetime.life");
            }
            WebTemplate_assign_int(W, "LDEF", life);
            WebTemplate_parse_dynamic(W, "reqfull.renew.lifetime");

            WebTemplate_parse_dynamic(W, "reqfull.renew");
         }
      }
      if (R->status == ST_REQ) {
         WebTemplate_parse_dynamic(W, "reqfull.cancel");
      }
      if (R->status == ST_DENY) {
         WebTemplate_parse_dynamic(W, "reqfull.cancel");
      }

      if (show_chks) {
         WebTemplate_parse_dynamic(W, "reqfull.rvnotice");
      } else if (showrev) {
         WebTemplate_parse_dynamic(W, "reqfull.revoke");
      }
      ret = 1;

   } else {
      char *emsg = (char*) malloc(128 + (errdn?strlen(errdn):0));
      if (errdn) sprintf(emsg, "Sorry, you are not an owner of request %d (<tt>%s</tt>)", reqno, errdn);
      else sprintf(emsg, "Sorry, request %d not found", reqno);
      WebTemplate_assign(W, "EMSG", emsg);
      free(emsg);
      ret = 0;
   }
   db_free_req();
   return (ret);
}


/* Check the domain ownership */

char **netact_get_host_ids(char *name);
char **netact_get_domain_ids(char *name);

int chk_dn_ownership(WebTemplate W, char *dn)
{
  char **list;
  int nl;
  char *dp, *edp;
  int nsd = 0;

  if ((!dn) || !(*dn)) return;

  /* Check for valid domain name */

  if (ssl_chk_dns(dn,0)) {
     WebTemplate_parse_dynamic(W, "form.nodns");

  /* Check netops */

  } else if (netact_test_name(dn, remote_user)) {
     WebTemplate_parse_dynamic(W, "form.dnok");
     return (1);

  /* Show who's got what */

  } else {

     WebTemplate_get_by_name(W, "dnhelp", "dnx.tpl");
     WebTemplate_assign(W, "SDN", dn);
     list = netact_get_host_ids(dn);
     if (list && *list) {
        nl = 0;
        while (*list) {
           WebTemplate_assign(W, "ID", *list);
           WebTemplate_parse_dynamic(W, "dnhelp.host.owners.id");
           list++;
           nl++;
        }
        if (nl==1) WebTemplate_parse_dynamic(W, "dnhelp.host.owners.one");
        else WebTemplate_parse_dynamic(W, "dnhelp.host.owners.many");
        WebTemplate_parse_dynamic(W, "dnhelp.host.owners");
     } else WebTemplate_parse_dynamic(W, "dnhelp.host.noowner");
     WebTemplate_parse_dynamic(W, "dnhelp.host");


     /* Show subdomains */

     if (edp=strstr(dn,".washington.edu")) {
        for (dp=strchr(dn,'.');dp&&dp<edp;dp=strchr(dp+1,'.')) {
          nsd = 0;
          WebTemplate_assign(W, "SDN", dp+1);
          list = netact_get_domain_ids(dp+1);
          if (list && *list) {
            nl = 0;
            while (*list) {
              WebTemplate_assign(W, "ID", *list);
              WebTemplate_parse_dynamic(W, "dnhelp.domain.owners.id");
              list++;
              nsd++;
              nl++;
            }
            if (nl==1) WebTemplate_parse_dynamic(W,
                           "dnhelp.domain.owners.one");
            else WebTemplate_parse_dynamic(W, "dnhelp.domain.owners.many");
            WebTemplate_parse_dynamic(W, "dnhelp.domain.owners");
          } else WebTemplate_parse_dynamic(W, "dnhelp.domain.noowner");
          WebTemplate_parse_dynamic(W, "dnhelp.domain");
        }
     }
     WebTemplate_parse(W, "DNHELP", "dnhelp");
  }
  return (0);
}


/* Process request */

main()
{
  WebTemplate W = newWebTemplate();
  char *userid;
  char *req;
  char *seqstr;
  int seq = 0;
  int s;
  char *srt;
  time_t now = time(NULL);
  int need_default = 1;
  int show_chks = 0;
  char *p;
  char *f1_pass = NULL;

  long rhash;
 
  signal(SIGPIPE, catch_pipe);

  WebTemplate_set_comments(W, "#", NULL);
  WebTemplate_add_header(W, "Cache-Control",
        "no-store, no-cache, must-revalidate");
  /* WebTemplate_add_header(W, "Expires", "Sat, 1 Jan 2000 01:01:01 GMT"); */
  WebTemplate_get_by_name(W, "page", "page.tpl");
  WebTemplate_get_by_name(W, "index", "index.tpl");

  /* Get arguments */

  WebTemplate_get_args(W);
  srt = set_sort(W, "uwcasrt_u");
  // WebTemplate_header(W);

  userid = remote_user;
  if (!userid) userid="(null)";
  seqstr = WebTemplate_get_arg(W, "seq");
  if (seqstr) seq = atoi(seqstr);
  req = WebTemplate_get_arg(W, "req");

  reqno = 0;
  p = WebTemplate_get_arg(W, "rno");
  if (p && *p) {
     reqno = atoi(p);
     if (reqno) WebTemplate_assign(W, "REQNO", p);
     else WebTemplate_assign(W, "REQNO", "-invalid-");
  }
  crtno = 0;
  p = WebTemplate_get_arg(W, "crt");
  if (p && *p) {
     crtno = atoi(p);
     if (crtno) WebTemplate_assign(W, "CRTNO", p);
     else WebTemplate_assign(W, "CRTNO", "-invalid-");
  }
     
  WebTemplate_assign(W, "TITLE", 
       "University of Washington Certificate Authority");
  WebTemplate_assign(W, "CURPAGE", "UW Services CA home");
  WebTemplate_assign(W, "ROOT", URLROOT);
  WebTemplate_assign(W, "CA", "UW Services CA");

  xsec_set_valid_form(W);

  WebTemplate_assign(W, "USER", userid);

  /* right side request selector */
  WebTemplate_parse_dynamic(W, "page.adminjs");
  WebTemplate_parse_dynamic(W, "page.services");

  /* Rest of page depends on request */

  crumb_init(W);


  /* the getder request sends the der data and exits */

  if (CREQ("getder")) {

     int rc;
     char *p;
     SvrCert S = (SvrCert) malloc(sizeof(SvrCert_));

     if (!db_connect()) show_db_error(W, "connect");
     
     db_select_crt(crtno, 0 ,'\0');

     if (db_get_crt(S)>0) {
        char *hdr = "Content-type: application/pkix-cert\n\n";
        char *der;
        int l;

        der = ssl_crttxt2der(S->cert, &l);
        if (der) {
           write(1, hdr, strlen(hdr));
           write(1, der, l);
           exit (0);
        }
     }
  }

  // We can now send the html content header
  WebTemplate_header(W);

  /* The first several requests cause an action then print
     one of the other pages. */

  /* Process a revoke request. 
     If the checkbox has been checked set the revocation status.
     Else show the checkbox.  */

  if (RREQ("revoke")) {

     int rc;
     char *p;
     char *revtxt;
     SvrReq R = (SvrReq) malloc(sizeof(SvrReq_));

     chk_xsec(W);

     revtxt = WebTemplate_get_arg(W, "revtxt");

     if (revtxt && *revtxt) {
        char smsg[64];

        /* set revoke */
        if (!db_connect()) show_db_error(W, "connect");
        db_select_req(reqno, NULL, userid, ST_NONE, NULL);

        if ((db_get_req(R)>0)) {
          db_set_req_status(reqno, ST_REVREQ, userid);
          db_load_notes(reqno, revtxt);
          // sprintf(smsg,"Revoke requested for request %d.", reqno);
          // WebTemplate_assign(W, "MSG", smsg);
          write_log(SERVICES_LOG, remote_user, "revreq", reqno, 0);

          show_action_status(W, R->reqno, "revoke_request");

        } else show_action_status(W, R->reqno, "not_yours");
        need_default = 0;

     } else {
        /* show warning */
        show_chks = 1;
        req = "full"; /* show list page */
     }
  }


  /* Process a completed request form.  */

  if (REQ("sub")) {

     int rc;
     char *treq;
     char *dn;
     char *cn;
     char *not;
     char *f1;
     char *err;
     int renew_req = 0;
     SvrReq R = (SvrReq) malloc(sizeof(SvrReq_));
     X509_REQ *xreq;
     FILE *errf = fopen(path_name("log/sub.log"), "a");
     time_t errt = time(NULL);
     SvrReq renR = (SvrReq) malloc(sizeof(SvrReq_));

     chk_xsec(W);

     need_default = 0;
     if (errf) fprintf(errf, "\n%s", ctime(&errt));
     if (errf) fflush(errf);

     if (!db_connect()) show_db_error(W, "connect");

     treq = WebTemplate_get_arg(W, "renew");
     if (treq && *treq) {
        renew_req = atoi(treq);
        WebTemplate_assign(W, "RENEW_REQNO", int2dec(renew_req));
        db_select_req(renew_req, NULL, NULL, ST_NONE, NULL);
        if ((db_get_req(renR)<=0)) {
           renR = NULL;
        }
        db_free_req();
     }

     treq = WebTemplate_get_arg(W, "reqtxt");
     if (!treq) treq =  "No certificate ";
     /* Add the 'begin-end' if not there */
     if (!strstr(treq,"CERTIFICATE")) {
        char *nr = (char*) malloc(strlen(treq)+98);
        strcat(strcat(strcpy(nr,"-----BEGIN CERTIFICATE REQUEST-----\n"),
                   treq),"\n-----END CERTIFICATE REQUEST-----\n");
        free(treq);
        treq = nr;
     }
     /* remove leading-trailing spaces */
     while (*treq==' ') treq++;
     s = strlen(treq);
     for (s=strlen(treq)-1;s>0;s--) {
        if ((treq[s]!=' ')&&(treq[s]!='\n')) break;
     }
     treq[s+1] = '\0';

     rhash = fnv_hash_32(treq);

     not = WebTemplate_get_arg(W, "nottxt");
     if (!not) not = "";
     f1 = WebTemplate_get_arg(W, "f1");
     if (!f1) f1 = "";

     if (errf) fprintf(errf, "hash=%d\nreq=%s\nnot=%s\n", rhash, treq, not);
     if (errf) fflush(errf);
     
     xreq = ssl_txt2req(treq);
     if (xreq) {
        if (errf) fprintf(errf, "req ok, dn=%s\n", ssl_get_dn_req(xreq));
        if (errf) fflush(errf);
     } else {
        if (errf) fprintf(errf, "req bogus\n");
        if (errf) fflush(errf);
     }

     if ((!xreq) || !(dn=ssl_get_dn_req(xreq))) {

        WebTemplate_assign(W,"EMSG",
             "We could not parse your form. Was that a certificate request?.");
        if (errf) fprintf(errf, "- parse failed\n");
        req = "wz5"; /* error page */

     } else if (renew_req && !renR) {
        char em[256];
        sprintf(em, "Could not find the request you are renewing: %d\n", renew_req);
        WebTemplate_assign(W,"EMSG",em);
        if (errf) fprintf(errf, "- no old request: %d\n", renew_req);
        req = "wz5";

     } else if (renew_req && strcmp(renR->dn, ssl_get_cn_req(xreq))) {
        char em[256];
        sprintf(em, "Renewing CN doesn't match: old=%d, %s\n", renew_req, renR->dn);
        WebTemplate_assign(W,"EMSG",em);
        req = "wz5";

     } else if (err=req_chk_dns(xreq)) {
        WebTemplate_assign(W,"EMSG", err);
        if (errf) fprintf(errf, "- invalid dns: %s\n",err);
        req = "wz5"; /* error page */
        
     } else if ( (renew_req==0) && ((rc=db_check_req_hash(rhash))>0) ) {

        char m[1024];

        sprintf(m, "We cannot process the request."
             "This request is a duplicate of request %d", rc);
        WebTemplate_assign(W,"EMSG",m);
        if (errf) fprintf(errf, "- duplicate\n");
        req = "wz5"; /* error page */ 

     } else {

        /* OK, have new request or ie renew requwst */

        char *ckemsg = NULL;
        int okdn;

     if (errf) fprintf(errf, "is OK - 1\n");
     if (errf) fflush(errf);
        if (!ckemsg) ckemsg = req_chk_policy(xreq, userid);
     if (errf) fprintf(errf, "is OK - 2, emsg=%s\n", ckemsg?ckemsg:"null");
     if (errf) fflush(errf);

        okdn = ckemsg==NULL;

        if (okdn || user_is_special(userid)) {
        // if (okdn) {

          char ssub[48];
          char *smsg;
          char mdt[4];
          char *nmsg = (char*) malloc(strlen(not)+512);
          char *nmsgp = nmsg;

     if (errf) fprintf(errf, "is OK - 2\n");
     if (errf) fflush(errf);
          *nmsg = '\0';
          strcpy(R->reqid, userid);
          R->status = ST_REQ;
          if (renew_req) R->status = ST_RENEW;
          R->request_date = time(NULL);
          R->crtno = 0;
          strcpy(R->dn, ssl_get_cn_req(xreq));
          strcpy(R->request, treq);


          if (reqno = db_load_req(R)) {
     if (errf) fprintf(errf, "loaded - reqno=%d\n", reqno);
     if (errf) fflush(errf);
             WebTemplate_assign(W, "REQNO", int2dec(reqno));
             WebTemplate_assign(W, "CN", R->dn);
             R->reqno = reqno;
             db_update_owners(R);

             /* If this is ie renew copy notes from old request */
             if (renew_req) {
                SvrNotes N = (SvrNotes) malloc(sizeof(SvrNotes_));
     if (errf) fprintf(errf, " - is renew\n");
     if (errf) fflush(errf);
                // copy notes - add "from old -> new
                db_select_notes(renew_req);
                while (db_get_notes(N)>0) {
                   db_load_notes(reqno, N->notes);
                }
                db_free_notes();
     if (errf) fprintf(errf, " - notes 0\n");
     if (errf) fflush(errf);
                smsg = (char*)malloc(256);
                sprintf(smsg, "Renewal of request %d by %s", renew_req, userid);
                db_load_notes(reqno, smsg);
                // also add to old ""renewed  as new"
                sprintf(smsg, "Renewed as request %d by %s", reqno, userid);
                db_load_notes(renew_req, smsg);
                free(smsg);
             }

             /* See if we have anything from the other forms - add to notes */
             if (f1=WebTemplate_get_arg(W, "f1")) {
                char *p1, *p2, *tc, *ts;
                p1=util_unpack(f1, "webnu");
                p2=util_unpack(f1, "webne");
     if (errf) fprintf(errf, " - notes 1\n");
     if (errf) fflush(errf);
                if ((p1&&*p1)||(p2&&*p2)) {
                   sprintf(nmsg, "Web Server: %s UW users, %s external\n",
                                    p1, p2);
                   while (*nmsgp) nmsgp++;
                }
                p1=util_unpack(f1, "tcl");
                p2=util_unpack(f1, "tsv");
                if (p1||p2) {
                   sprintf(nmsgp, "%s %s\n", p1?"SSL Client":"",
                                           p2?"SSL Server":"");
                   while (*nmsgp) nmsgp++;
                }
                p1=util_unpack(f1, "purp");
                if (p1) strcpy(nmsgp, p1);
                if (*nmsg) db_load_notes(reqno, nmsg);
          
                /* See if there's a request lifetime */
                if (p1 = util_unpack(f1, "days")) {
                   int l = atoi(p1);
     if (errf) fprintf(errf, " - days %d\n", l);
     if (errf) fflush(errf);
                   if (l>0 && l<=MAX_CERT_LIFE && l!=DEF_CERT_LIFE) {
                      sprintf(nmsg, "Requested lifetime: %d month%c.", l, l>1?'s':'\0');
                      db_load_notes(reqno, nmsg);
                      if (errf) fprintf(errf, "- life: %d\n", l);
                   }
                }
                
                free(f1);
             }

     if (errf) fprintf(errf, "- OK\n");
     if (errf) fflush(errf);
          } else {
             show_db_error(W, "loading req");
             if (errf) fprintf(errf, "- load req failed\n");
             if (errf) fflush(errf);
          }
          db_free_req();

          /* send out a notice */

     if (errf) fprintf(errf, "- sending notice\n");
     if (errf) fflush(errf);
          strcpy(mdt, "ok");
          if (!okdn) {
             if (ckemsg) strncpy(mdt, ckemsg, 2); /* note same size */
             else strcpy(mdt, "no"); 
          }
          sprintf(ssub,"%s %d (%s) from %s", renew_req?"Renew":"Req", reqno, mdt, remote_user);

          smsg = (char*)malloc(256 + strlen(R->dn) + strlen(nmsg));
          sprintf(smsg,"cn = %s\n\n%s\n\n" ADMIN_URL "\n\n", R->dn, nmsg); 

          if (strcmp(remote_user,"fox")) send_mail(GETS_MAIL, ssub, smsg);
          free(smsg);

          write_log(SERVICES_LOG, remote_user, "req", reqno, 0);

          /* If not compliant, send the user a notice */

          if ((!okdn) && (smsg=get_file(BADDN_FILE, NULL))) {
            sprintf(ssub,"%s@u.washington.edu", R->reqid);
            send_mail(ssub, R->dn, smsg);
          }

          show_action_status(W, R->reqno, renew_req?"ierenew_request":"new_request");
          need_default = 0;

        } else {

           char *em = (char*)malloc(strlen(ckemsg)+128);
           sprintf(em, "We cannot process the request.<br>%s", ckemsg+4);
           if (!strncmp(ckemsg,"AN",2)) strcat(em, " (subjectAltName)");
           WebTemplate_assign(W,"EMSG", em);
           if (errf) fprintf(errf, "- invalid cn: %s\n",ssl_get_cn_req(xreq));
           req = "wz4"; /* send the form again */
        }
     }
     if (errf) fclose(errf);

  }


  /* Process a rewewal request.

     If Windows, send a new request page
     else just update the status so the cert will be resigned. */

  if (RREQ("renew")) {

     int rc;
     char *p, *p1;
     SvrReq R = (SvrReq) malloc(sizeof(SvrReq_));
     SvrCert S = (SvrCert) malloc(sizeof(SvrCert_));

     chk_xsec(W);
     need_default = 0;

     if (!db_connect()) show_db_error(W, "connect");
     
     // db_select_req(reqno, NULL, userid, ST_NONE, NULL);
     db_select_req(reqno, NULL, NULL, ST_NONE, NULL);

        if ((db_get_req(R)>0)) {
          char ssub[48];
          char *smsg;
          char nmsg[64];
          db_free_req();

          /* if lifetime set, save it now. */
          if (p=WebTemplate_get_arg(W, "days")) {
             char *pd = WebTemplate_get_arg(W, "defdays");
             int l = atoi(p);
             int ld = 0;
             if (pd) ld = atoi(pd);
             if (l>0 && l<=MAX_CERT_LIFE && l!=ld) {
                sprintf(nmsg, "Requested lifetime: %d month%c.", l, l>1?'s':'\0');
                db_load_notes(reqno, nmsg);
             }
          }
          
          if (is_windows_req(R)) {
             char *sn;

             /* windows request.  Send out the last part of
                the normal cert request sequence */
  
             db_select_crt(R->crtno, R->reqno, 0);
             if (!db_get_crt(S)) { 
                // error - no cert
             }
                
             WebTemplate_assign(W, "PAGETITLE", "Renew certificate");
             add_crumb(W, "services/?req=list", "Manage certificates");
             add_req_crumb(W, R->reqno);

             WebTemplate_assign(W, "CN", R->dn);
             WebTemplate_assign(W, "CASN", int2hex(S->casn)+2); /* skip the "0x" */
             WebTemplate_get_by_name(W, "iehead", "ms/iehead.tpl");
             // WebTemplate_parse_dynamic(W, "iehead.capicom");
             WebTemplate_get_by_name(W, "iebody", "ms/iebody.tpl");
             WebTemplate_get_by_name(W, "form", "ms/ierenew.tpl");
           
             WebTemplate_parse(W, "HEADTXT", "iehead");
             WebTemplate_parse(W, "BODYTXT", "iebody");
             WebTemplate_parse(W, "BODY", "form");

          } else {

             /* non-windows request */

             rc = db_set_req_status(R->reqno, ST_RENEW, userid);
             if (!rc) show_db_error(W, "set rs");

             WebTemplate_assign(W, "CN", R->dn);

             /* send out a notice */
             sprintf(ssub,"renew %d from %s", reqno, remote_user); 
             smsg = (char*)malloc(256 + strlen(R->dn));
             sprintf(smsg,"cn = %s\n\n" ADMIN_URL "\n\n", R->dn); 
             if (strcmp(remote_user,"fox")) send_mail(GETS_MAIL, ssub, smsg);
             free(smsg);

             write_log(SERVICES_LOG, remote_user, "renew", reqno, 0);

             show_action_status(W, R->reqno, "renew_request");
           }
        
        } else show_action_status(W, R->reqno, "not_yours");
        need_default = 0;

  }


  /* Process a cancellation request */

  if (RREQ("cancel")) {

     int rc;
     char *p;
     SvrReq R = (SvrReq) malloc(sizeof(SvrReq_));
     
     chk_xsec(W);

     need_default = 0;

     if (!db_connect()) show_db_error(W, "connect");
     
     db_select_req(reqno, NULL, userid , ST_REQ, NULL);

     if ((db_get_req(R)>0)) {
       char ssub[48];
       db_free_req();
       rc = db_delete_req(R->reqno);
       if (!rc) show_db_error(W, "del req");
       else {
          WebTemplate_assign(W, "CN", R->dn);

          /* send out a notice */
          sprintf(ssub,"request %d cancelled", reqno); 
          if (strcmp(remote_user,"fox")) send_mail(GETS_MAIL, ssub,
             "So you can forget about it.\n");

          write_log(SERVICES_LOG, remote_user, "cancel", reqno, 0);
       }
       show_action_status(W, R->reqno, "cancel_request");

     } else show_action_status(W, R->reqno, "not_yours");
     need_default = 0;
  }


  /* The following requests grenerate one of the displays */


  /* The request wizard sequence:

       req    Page to show                   template
       ----   -------------------------      --------------

       "wz1"  Cautionary notice              "wz_1.tpl"
       "wz1b" Host ownership                 "wz_1b.tpl"
       "wz2"  Purpose form                   "wz_2.tpl"
       "wz3"  Method selection               "wz_3.tpl"
       "wz4"  PEM form                       "wz_4.tpl"
       "wz5"  Error from submit              "wz_5.tpl"

       Any inputs from a form are packed into "f1".

   */

           

  /* Receive the 'purpose' form and send the 'method'
     If essential parts of the form are not filled - go back */

  if (REQ("wz3")) {

     char *f1 = NULL;
     int oldf1 = 1;
     char *pd = NULL;
     char *p0, *p1, *p2, *p3, *p4;
     char *pn, *po;
     
     chk_xsec(W);

     /* Recover form vars */

     if (!(f1=WebTemplate_get_arg(W, "f1"))) {
       oldf1 = 0;
       if (p0=WebTemplate_get_arg(W, "webck")) util_pack(&f1, "webck", p0);
       if (p1=WebTemplate_get_arg(W, "webnu")) util_pack(&f1, "webnu", p1);
       if (p2=WebTemplate_get_arg(W, "webne")) util_pack(&f1, "webne", p2);
       if (p3=WebTemplate_get_arg(W, "tcl")) util_pack(&f1, "tcl", p3);
       if (p4=WebTemplate_get_arg(W, "tsv")) util_pack(&f1, "tsv", p4);
       if (pd=WebTemplate_get_arg(W, "purp")) util_pack(&f1, "purp", pd);
       if (pn=WebTemplate_get_arg(W, "dn")) util_pack(&f1, "dn", pn);
       if (po=WebTemplate_get_arg(W, "owner")) util_pack(&f1, "owner", po);
       if (po=WebTemplate_get_arg(W, "days")) util_pack(&f1, "days", po);
     }
     WebTemplate_assign(W, "F1", f1);
     
     if (oldf1 || (has_some_text(pd) && 
          (has_some_text(p1)||has_some_text(p2)||p3||p4))) {
       
        Browser B = get_browser();

        WebTemplate_parse_dynamic(W, "page.formjs");

        add_back_crumb(W, "wz1", "Notice");
        add_back_crumb(W, "wz1b", "Host");
        add_back_crumb(W, "wz2", "Purpose");

        WebTemplate_assign(W, "PAGETITLE", "Certificate request - Method");
        WebTemplate_get_by_name(W, "form", "wz_3.tpl");
     
        if (B->type!=BR_IE) {
            WebTemplate_parse_dynamic(W, "form.noie");
            WebTemplate_assign(W, "IEDISABLED", "disabled");
        }
        WebTemplate_parse(W, "BODY", "form");
        free (f1);
        need_default = 0;

        if (pn) WebTemplate_assign(W, "DN", pn);
        WebTemplate_parse_dynamic(W, "page.request_cn");
        

     } else { /* have to go back */
        
        if (p0 && !(has_some_text(p1)&&has_some_text(p2))) {
           WebTemplate_assign(W, "EMSG", "Please indicate an estimated user count.");
        } else if (!(p0||p3||p4)) {
           WebTemplate_assign(W, "EMSG", "Please select at least one purpose.");
        } else if (!(pd&&*(pd))) {
           WebTemplate_assign(W, "EMSG", "Please briefly describe your application.");
        } else {
           /* should not get here */
           WebTemplate_assign(W, "EMSG", "Please complete the form.");
        }
        req = "wz2";
        f1_pass = f1;
     }

  }


  /* Send the cert wizard - part1 - the warning notice */

  if (REQ("wz1")) {

     WebTemplate_assign(W, "PAGETITLE", "Certificate request - Notice");
     WebTemplate_get_by_name(W, "form", "wz_1.tpl");
     
     /* crumb_help(W, "services/?req=formh"); */

     WebTemplate_parse(W, "BODY", "form");


  /* Send the cert wizard - part1b - the domain notice */

  } else if (REQ("wz1b")) {

     char *ckdn;

     WebTemplate_assign(W, "PAGETITLE",
              "Certificate request - Host ownership");
     WebTemplate_get_by_name(W, "form", "wz_1b.tpl");
     WebTemplate_parse_dynamic(W, "page.formjs");
     
     ckdn = WebTemplate_get_arg(W, "ckdn");
     if (ckdn && *ckdn) {
        fix_dns_name(ckdn);
        WebTemplate_assign(W, "CKDN", ckdn);
        if (chk_dn_ownership(W, ckdn)) WebTemplate_assign(W, "DN", ckdn);
     }

     add_back_crumb(W, "wz1", "Notice");
     
     /* crumb_help(W, "services/?req=formh"); */

     WebTemplate_parse(W, "BODY", "form");


  /* Send the cert wizard - part2  purpose */

  } else if (REQ("wz2")) {

     char *f1, *dn;
     SvrUG_ G;
     CertLife cl;
     int pdays=DEF_CERT_LIFE;

     chk_xsec(W);

     if ((f1=f1_pass) || (f1=WebTemplate_get_arg(W, "f1"))) {
        char *p;
        if (p=util_unpack(f1, "webnu")) WebTemplate_assign(W, "WEBNU", p);
        if (p=util_unpack(f1, "webne")) WebTemplate_assign(W, "WEBNE", p);
        if (util_unpack(f1, "webck")) WebTemplate_assign(W, "WEBCK", CHECK);
        if (util_unpack(f1, "tcl")) WebTemplate_assign(W, "TCLCK", CHECK);
        if (util_unpack(f1, "tsv")) WebTemplate_assign(W, "TSVCK", CHECK);
        if (p=util_unpack(f1, "purp")) WebTemplate_assign(W, "PURP", p);
        p = util_unpack(f1, "days");
        if (p) pdays = atoi(p);
     }
     if (pdays<0 || pdays>MAX_CERT_LIFE) pdays = DEF_CERT_LIFE;

     if (dn=WebTemplate_get_arg(W, "dn")) {
        WebTemplate_assign(W, "DN", dn);
        WebTemplate_parse_dynamic(W, "page.request_cn");
     }

     WebTemplate_parse_dynamic(W, "page.formjs");
     add_back_crumb(W, "wz1", "Notice");
     add_back_crumb(W, "wz1b", "Host");

     WebTemplate_assign(W, "PAGETITLE", "Certificate request - Purpose");
     WebTemplate_get_by_name(W, "form", "wz_2.tpl");

     /* add days selector */
     for (cl=cert_lifetimes; cl->mon; cl++) {
        WebTemplate_assign_int(W, "LMON", cl->mon);
        WebTemplate_assign(W, "LTXT", cl->txt);
        WebTemplate_assign(W, "LSEL", (pdays==cl->mon)?SELECT:"");
        WebTemplate_parse_dynamic(W, "form.lifetime.life");
     }
     WebTemplate_parse_dynamic(W, "form.lifetime");
     
     /* crumb_help(W, "services/?req=formh"); */

     WebTemplate_parse(W, "BODY", "form");



  /* Receive the method - send the pem */

  } else if (REQ("wz4")) {

     char *f1;
     char *dn = NULL;
     FILE *errf = fopen(path_name("log/sub.log"), "a");

     chk_xsec(W);

     if (f1=WebTemplate_get_arg(W, "f1")) {
        WebTemplate_assign(W, "F1", f1);
        dn = util_unpack(f1, "dn");
        free(f1);
     }
     if (dn&&*dn) {
        WebTemplate_assign(W, "DN", dn);
        WebTemplate_parse_dynamic(W, "page.request_cn");
     } else WebTemplate_assign(W, "DISD", "disabled");

     WebTemplate_parse_dynamic(W, "page.formjs");
     add_back_crumb(W, "wz1", "Notice");
     add_back_crumb(W, "wz1b", "Host");
     add_back_crumb(W, "wz2", "Purpose");
     add_back_crumb(W, "wz3", "Method");

     WebTemplate_assign(W, "PAGETITLE", "Certificate request - Submit PEM");

     if (WebTemplate_get_arg(W, "form4ms")) {    /* ActiveX method */

        WebTemplate_assign(W, "PAGETITLE", "ActiveX certificate request");
        WebTemplate_get_by_name(W, "iehead", "ms/iehead.tpl");
        WebTemplate_get_by_name(W, "iebody", "ms/iebody.tpl");
        WebTemplate_get_by_name(W, "form", "ms/iereq.tpl");
     
        WebTemplate_parse(W, "HEADTXT", "iehead");
        WebTemplate_parse(W, "BODYTXT", "iebody");
        WebTemplate_parse(W, "BODY", "form");

     } else {
        
       if (dn&&*dn) {
          WebTemplate_assign(W, "DN", dn);
          WebTemplate_parse_dynamic(W, "page.request_cn");
       } 
       WebTemplate_get_by_name(W, "form", "wz_4.tpl");
       if (WebTemplate_get_arg(W, "form4pu")) {
          WebTemplate_parse_dynamic(W, "form.unix");
       } else {
          WebTemplate_parse_dynamic(W, "form.windows");
       }
       /* crumb_help(W, "services/?req=formh"); */

     }

     WebTemplate_parse(W, "BODY", "form");


  /* Error during process of submit - from other commands*/

  } else if (REQ("wz5")) {

     char *f1;
     char *dn = NULL;

     WebTemplate_assign(W, "PAGETITLE", "Certificate request - Error");

     WebTemplate_get_by_name(W, "error", "wz_5.tpl");
     add_crumb(W, "services/?req=list", "Manage certificates");
     if (reqno) add_req_crumb(W, reqno);
     /* crumb_help(W, "services/?req=formh"); */

     WebTemplate_parse(W, "BODY", "form");



  /* Send the PEM cert form help text */

  } else if (REQ("formh")) {

     add_crumb(W, "services/?req=form", "Cert Request");
     WebTemplate_assign(W, "PAGETITLE", "Requesting a certificate");
     WebTemplate_get_by_name(W, "help", "help_req.tpl");
     
     WebTemplate_parse(W, "BODY", "help");


  /* Get a particular cert */

  } else if (CREQ("get")) {

     int rc;
     char pt[64];
     char pl[64];
     SvrCert S = (SvrCert) malloc(sizeof(SvrCert_));
     SvrReq R = (SvrReq) malloc(sizeof(SvrReq_));

     add_crumb(W, "services/?req=list", "Manage certificates");
     sprintf(pl, "services/?req=full&rno=%d", reqno);
     sprintf(pt, "Request %d", reqno);
     add_crumb(W, pl, pt);
     sprintf(pt, "Certificate for request %d", reqno);

     WebTemplate_assign(W, "PAGETITLE", pt);
     WebTemplate_get_by_name(W, "crttop", "reqcrt_top.tpl");
     WebTemplate_get_by_name(W, "crtbody", "pem_body.tpl");

     if (!db_connect()) show_db_error(W, "connect");
     
     db_select_crt(crtno, 0 ,'\0');

     if (db_get_crt(S)>0) {
        db_select_req(S->reqno, NULL, NULL, RST_ALIVE, NULL);
        if ((db_get_req(R)>0)) {
          if (R->status==ST_RENEW) WebTemplate_assign(W, "MSG",
              "This certificate has not yet been renewed!");
          WebTemplate_assign(W, "PEM", S->cert);
          WebTemplate_assign(W, "CN", R->dn);
          WebTemplate_assign(W, "SDATE", date2mdy(S->issue_date));
          WebTemplate_assign(W, "EDATE", date2mdy(S->expire_date));
          if (is_windows_req(R)) {
             Browser B = get_browser();
             if (B->type==BR_IE) WebTemplate_parse_dynamic(W, "crttop.getms");
             else WebTemplate_parse_dynamic(W, "crttop.getmsnot");
          }
        } else WebTemplate_assign(W, "MSG", "Not your certificate!");
     }

     sprintf(pl, "services/?req=geth&rno=%d&crt=%d", reqno, crtno);
     crumb_help(W, pl);

     WebTemplate_parse(W, "BODY", "crtbody");
     WebTemplate_parse(W, "TOPBODY", "crttop");


  /* Send help text for the view page */

  } else if (REQ("geth")) {

     char pt[64];
     char pl[64];

     add_crumb(W, "services/?req=list", "Manage certificates");

     sprintf(pl, "services/?req=full&rno=%d", reqno);
     sprintf(pt, "Request %d", reqno);
     add_crumb(W, pl, pt);

     sprintf(pl, "services/?req=get&rno=%d&crt=%d", reqno, crtno);
     sprintf(pt, "Get %d", reqno);
     add_crumb(W, pl, pt);

     WebTemplate_assign(W, "PAGETITLE", "Retrieving a certificate");
     WebTemplate_get_by_name(W, "help", "help_get.tpl");
     
     WebTemplate_parse(W, "BODY", "help");



  /* Get a particular cert (DER format) */

  } else if (CREQ("getder")) {

     int rc;
     char *p;
     SvrCert S = (SvrCert) malloc(sizeof(SvrCert_));

     if (!db_connect()) show_db_error(W, "connect");
     
     db_select_crt(crtno, 0 ,'\0');

     if (db_get_crt(S)>0) {
        char *hdr = "Content-type: application/pkix-cert\n\n";
        char *der;
        int l;

        der = ssl_crttxt2der(S->cert, &l);
        if (der) {
           write(1, hdr, strlen(hdr));
           write(1, der, l);
           exit (0);
        }
     }


  /* Get a particular cert (ActiveX method) */

  } else if (CREQ("getms")) {

     int rc;
     char *p;
     SvrCert S = (SvrCert) malloc(sizeof(SvrCert_));
     SvrReq R = (SvrReq) malloc(sizeof(SvrReq_));
     SvrCert T = (SvrCert) malloc(sizeof(SvrCert_));
     int lsn = 0;

     chk_xsec(W);

     WebTemplate_assign(W, "CURPAGE", "Certificate request");
     WebTemplate_assign(W, "RETURN", "/services/?req=list");
     WebTemplate_get_by_name(W, "iehead", "ms/iehead.tpl");
     WebTemplate_get_by_name(W, "iebody", "ms/iebody.tpl");
     WebTemplate_get_by_name(W, "loadtop", "ms/ieload.tpl");
     
     if (!db_connect()) show_db_error(W, "connect");
     
     db_select_crt(crtno, 0 ,'\0');

     if (db_get_crt(S)>0) {
        WebTemplate_assign(W, "VBCERT", fix_cert_for_vbs(S->cert));
        WebTemplate_assign(W, "CASN", int2hex(S->casn)+2);
     } else WebTemplate_assign(W, "MSG", "Cert not found!");
     db_free_crt();

     db_select_req(reqno, NULL, NULL, ST_NONE, NULL);
     if (db_get_req(R)>0) {
        WebTemplate_assign(W, "CN", R->dn);
     }
     db_free_req();
     
     add_crumb(W, "services/?req=list", "Manage certificates");
     add_req_crumb(W, reqno);

     WebTemplate_parse(W, "HEADTXT", "iehead");
     WebTemplate_parse(W, "BODYTXT", "iebody");
     WebTemplate_parse(W, "TOPBODY", "loadtop");
     WebTemplate_assign(W, "BODY", "");


  /* Process a detail (full) request. */

  } else if (RREQ("full")) {

     int rc;
     char pt[64];

     add_crumb(W, "services/?req=list", "Manage certificates");

     sprintf(pt, "Request %d", reqno);

     WebTemplate_assign(W, "PAGETITLE", pt);
     WebTemplate_get_by_name(W, "reqfull", "reqfull.tpl");

     if (reqno && show_req_detail(W, reqno, show_chks)) {
        WebTemplate_parse(W, "DETAIL", "reqfull");
     }
     

  /* Show current requests */

  } else if (REQ("list") || need_default) {

     int rc;
     char pt[64];
     SvrReq R = (SvrReq) malloc(sizeof(SvrReq_));
     SvrCert S = (SvrCert) malloc(sizeof(SvrCert_));
     SvrNotes N = (SvrNotes) malloc(sizeof(SvrNotes_));

     sprintf(pt, "Requests and Certificates for %s", remote_user);
     WebTemplate_assign(W, "PAGETITLE", pt);
     WebTemplate_get_by_name(W, "list", "reqlist.tpl");
     if (srt && srt[1]=='n') WebTemplate_parse_dynamic(W, "list.show_rev");
     else WebTemplate_parse_dynamic(W, "list.hide_rev");

     if (!db_connect()) show_db_error(W, "connect");
     
     db_select_req(0, NULL, userid, ST_NONE, srt);

     while (db_get_req(R)>0) {

       if (R->status == ST_DELETED) continue;
       if (R->status == ST_CANCEL) continue;
       if (R->status == ST_EXPIRED) continue;

       WebTemplate_assign(W, "REQNO", int2dec(R->reqno));
       WebTemplate_assign(W, "CN", R->dn);
       WebTemplate_assign(W, "RDATE", date2mdy(R->request_date));
       WebTemplate_assign(W, "OWNER", R->reqid);
       WebTemplate_assign(W, "STATUS", req_status_text[R->status]);
       WebTemplate_assign(W, "EDATE", "&nbsp;"); /* in case no crt */

       if (R->status==ST_ISSUE) WebTemplate_assign(W, "REQCS", "reqok");
       else if (R->status==ST_REQ) WebTemplate_assign(W, "REQCS", "reqip");
       else if (R->status==ST_RENEW) WebTemplate_assign(W, "REQCS", "reqip");
       else WebTemplate_assign(W, "REQCS", "reqno");

       if (R->crtno) {
         db_select_crt(R->crtno, 0, 0);
         if (db_get_crt(S)) {
            WebTemplate_assign(W, "SDATE", date2mdy(S->issue_date));
            WebTemplate_assign(W, "EDATE", date2mdy(S->expire_date));
            if ((S->status==CRT_EXPIRING)||(S->expire_date<now+MONTH)) {
               char st[256];
               int nd = (S->expire_date-now)/DAY;
               if (nd>=0) sprintf(st, "%s<br>(%d day%c)", date2mdy(S->expire_date), nd, nd==1?'\0':'s');
               else sprintf(st, "%s<br>(expired)", date2mdy(S->expire_date));
               WebTemplate_assign(W, "EDATE", st);
               WebTemplate_assign(W, "REQCS", "reqex");
            }
         }
         db_free_crt();
       }

       /* Any notes? */

       db_select_notes(R->reqno);
       while (db_get_notes(N)>0) {
          WebTemplate_assign(W, "NOTE", fix_string_for_html(N->notes)); 
          /* WebTemplate_assign(W, "NOTE", N->notes); */
          WebTemplate_parse_dynamic(W, "list.req.notes");
       }
       /* db_free_notes(); */
       WebTemplate_parse_dynamic(W, "list.req");
       
     }
     WebTemplate_parse(W, "BODY", "list");


  } else if (need_default) {
     WebTemplate_parse(W, "BODY", "index"); 
     /* WebTemplate_assign(W, "BODY",
          "<meta http-equiv=\"Refresh\" content=\"1;url=/\""); */
  }
 
  WebTemplate_parse(W, "PAGE", "page");
  WebTemplate_write(W, "PAGE");

}

